from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Sequence
from typing import Generic, Literal, cast
from uuid import uuid4

from cognite.client.data_classes.data_modeling import (
    NodeId,
    ViewId,
)
from cognite.client.exceptions import CogniteException

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.resource_classes.chart import ChartRequest, ChartResponse
from cognite_toolkit._cdf_tk.client.resource_classes.charts_data import (
    ChartCoreTimeseries,
    ChartSource,
    ChartTimeseries,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    EdgeRequest,
    InstanceRequest,
    NodeReference,
    NodeRequest,
    ViewReference,
    ViewResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import InstanceIdentifier, TypedNodeIdentifier
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.canvas import (
    ContainerReferenceApply,
    FdmInstanceContainerReferenceApply,
    IndustrialCanvas,
    IndustrialCanvasApply,
)
from cognite_toolkit._cdf_tk.client.resource_classes.resource_view_mapping import (
    RESOURCE_VIEW_MAPPING_SPACE,
    ResourceViewMappingRequest,
)
from cognite_toolkit._cdf_tk.client.resource_classes.three_d import (
    AssetMappingClassicResponse,
    AssetMappingDMRequest,
    RevisionStatus,
    ThreeDModelClassicResponse,
)
from cognite_toolkit._cdf_tk.commands._migrate.conversion import DirectRelationCache, asset_centric_to_dm
from cognite_toolkit._cdf_tk.commands._migrate.data_classes import (
    Model,
    ThreeDMigrationRequest,
    ThreeDRevisionMigrationRequest,
)
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import create_default_mappings
from cognite_toolkit._cdf_tk.commands._migrate.issues import (
    CanvasMigrationIssue,
    ChartMigrationIssue,
    ConversionIssue,
    ThreeDModelMigrationIssue,
)
from cognite_toolkit._cdf_tk.constants import MISSING_INSTANCE_SPACE
from cognite_toolkit._cdf_tk.exceptions import ToolkitMigrationError, ToolkitValueError
from cognite_toolkit._cdf_tk.protocols import T_ResourceRequest, T_ResourceResponse
from cognite_toolkit._cdf_tk.storageio._base import T_Selector
from cognite_toolkit._cdf_tk.storageio.logger import DataLogger, NoOpLogger
from cognite_toolkit._cdf_tk.storageio.selectors import CanvasSelector, ChartSelector, ThreeDSelector
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.useful_types2 import T_AssetCentricResourceExtended

from .data_classes import AssetCentricMapping
from .selectors import AssetCentricMigrationSelector


class DataMapper(Generic[T_Selector, T_ResourceResponse, T_ResourceRequest], ABC):
    def __init__(self, client: ToolkitClient) -> None:
        self.client = client
        self.logger: DataLogger = NoOpLogger()

    def prepare(self, source_selector: T_Selector) -> None:
        """Prepare the data mapper with the given source selector.

        Args:
            source_selector: The selector for the source data.

        """
        # Override in subclass if needed.
        pass

    @abstractmethod
    def map(self, source: Sequence[T_ResourceResponse]) -> Sequence[T_ResourceRequest | None]:
        """Map a chunk of source data to the target format.

        Args:
            source: The source data chunk to be mapped.

        Returns:
            A sequence of mapped target data.

        """
        raise NotImplementedError("Subclasses must implement this method.")


class AssetCentricMapper(
    DataMapper[AssetCentricMigrationSelector, AssetCentricMapping[T_AssetCentricResourceExtended], InstanceRequest]
):
    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._ingestion_view_by_id: dict[ViewReference, ViewResponse] = {}
        self._view_mapping_by_id: dict[str, ResourceViewMappingRequest] = {}
        self._direct_relation_cache = DirectRelationCache(client)

    def prepare(self, source_selector: AssetCentricMigrationSelector) -> None:
        ingestion_view_ids = source_selector.get_ingestion_mappings()
        node_ids = TypedNodeIdentifier.from_str_ids(ingestion_view_ids, space=RESOURCE_VIEW_MAPPING_SPACE)
        ingestion_views = self.client.migration.resource_view_mapping.retrieve(node_ids)
        defaults = {mapping.external_id: mapping for mapping in create_default_mappings()}
        # Custom mappings from CDF override the default mappings
        self._view_mapping_by_id = defaults | {view.external_id: view.as_request_resource() for view in ingestion_views}
        missing_mappings = set(ingestion_view_ids) - set(self._view_mapping_by_id.keys())
        if missing_mappings:
            raise ToolkitValueError(
                f"The following ingestion views were not found: {humanize_collection(missing_mappings)}"
            )

        view_ids = list(
            {
                ViewReference(
                    space=mapping.view_id.space,
                    external_id=mapping.view_id.external_id,
                    version=mapping.view_id.version,
                )
                for mapping in self._view_mapping_by_id.values()
            }
        )
        views = self.client.tool.views.retrieve(view_ids)
        self._ingestion_view_by_id = {view.as_id(): view for view in views}
        missing_views = set(view_ids) - set(self._ingestion_view_by_id.keys())
        if missing_views:
            raise ToolkitValueError(
                f"The following ingestion views were not found in Data Modeling: {humanize_collection(missing_views)}"
            )

    def map(
        self, source: Sequence[AssetCentricMapping[T_AssetCentricResourceExtended]]
    ) -> Sequence[InstanceRequest | None]:
        """Map a chunk of asset-centric data to InstanceApplyList format."""
        # We update the direct relation cache in bulk for all resources in the chunk.
        self._direct_relation_cache.update(item.resource for item in source)
        output: list[InstanceRequest | None] = []
        issues: list[ConversionIssue] = []
        for item in source:
            instance, conversion_issue = self._map_single_item(item)
            identifier = str(item.mapping.as_asset_centric_id())

            if conversion_issue.missing_instance_space:
                self.logger.tracker.add_issue(identifier, "Missing instance space")
            if conversion_issue.failed_conversions:
                self.logger.tracker.add_issue(identifier, "Failed conversions")
            if conversion_issue.invalid_instance_property_types:
                self.logger.tracker.add_issue(identifier, "Invalid instance property types")
            if conversion_issue.missing_asset_centric_properties:
                self.logger.tracker.add_issue(identifier, "Missing asset-centric properties")
            if conversion_issue.missing_instance_properties:
                self.logger.tracker.add_issue(identifier, "Missing instance properties")
            if conversion_issue.ignored_asset_centric_properties:
                self.logger.tracker.add_issue(identifier, "Ignored asset-centric properties")

            if conversion_issue.has_issues:
                issues.append(conversion_issue)

            if instance is None:
                self.logger.tracker.finalize_item(identifier, "failure")
            output.append(instance)
        if issues:
            self.logger.log(issues)
        return output

    def _map_single_item(
        self, item: AssetCentricMapping[T_AssetCentricResourceExtended]
    ) -> tuple[NodeRequest | EdgeRequest | None, ConversionIssue]:
        mapping = item.mapping
        ingestion_view = mapping.get_ingestion_view()
        try:
            view_source = self._view_mapping_by_id[ingestion_view]
            view_ref = ViewReference(
                space=view_source.view_id.space,
                external_id=view_source.view_id.external_id,
                version=view_source.view_id.version,
            )
            view_properties = self._ingestion_view_by_id[view_ref].properties
        except KeyError as e:
            raise RuntimeError(
                f"Failed to lookup mapping or view for ingestion view '{ingestion_view}'. Did you forget to call .prepare()?"
            ) from e
        instance, conversion_issue = asset_centric_to_dm(
            item.resource,
            instance_id=mapping.instance_id,
            view_source=view_source,
            view_properties=view_properties,
            direct_relation_cache=self._direct_relation_cache,
            preferred_consumer_view=mapping.preferred_consumer_view,
        )
        if mapping.instance_id.space == MISSING_INSTANCE_SPACE:
            conversion_issue.missing_instance_space = f"Missing instance space for dataset ID {mapping.data_set_id!r}"
        return instance, conversion_issue


class ChartMapper(DataMapper[ChartSelector, ChartResponse, ChartRequest]):
    def map(self, source: Sequence[ChartResponse]) -> Sequence[ChartRequest | None]:
        self._populate_cache(source)
        output: list[ChartRequest | None] = []
        issues: list[ChartMigrationIssue] = []
        for item in source:
            mapped_item, issue = self._map_single_item(item)
            identifier = item.external_id

            if issue.missing_timeseries_ids:
                self.logger.tracker.add_issue(identifier, "Missing timeseries IDs")
            if issue.missing_timeseries_external_ids:
                self.logger.tracker.add_issue(identifier, "Missing timeseries external IDs")
            if issue.missing_timeseries_identifier:
                self.logger.tracker.add_issue(identifier, "Missing timeseries identifier")

            if issue.has_issues:
                issues.append(issue)

            if mapped_item is None:
                self.logger.tracker.finalize_item(identifier, "failure")
            output.append(mapped_item)
        if issues:
            self.logger.log(issues)
        return output

    def _populate_cache(self, source: Sequence[ChartResponse]) -> None:
        """Populate the internal cache with timeseries from the source charts.

        Note that the consumption views are also cached as part of the timeseries lookup.
        """
        timeseries_ids: set[int] = set()
        timeseries_external_ids: set[str] = set()
        for chart in source:
            for item in chart.data.time_series_collection or []:
                if item.ts_id:
                    timeseries_ids.add(item.ts_id)
                if item.ts_external_id:
                    timeseries_external_ids.add(item.ts_external_id)
        if timeseries_ids:
            self.client.migration.lookup.time_series(list(timeseries_ids))
        if timeseries_external_ids:
            self.client.migration.lookup.time_series(external_id=list(timeseries_external_ids))

    def _map_single_item(self, item: ChartResponse) -> tuple[ChartRequest | None, ChartMigrationIssue]:
        issue = ChartMigrationIssue(chart_external_id=item.external_id, id=item.external_id)
        time_series_collection = item.data.time_series_collection or []
        timeseries_core_collection = self._create_timeseries_core_collection(time_series_collection, issue)
        if issue.has_issues:
            return None, issue

        updated_source_collection = self._update_source_collection(
            item.data.source_collection or [], time_series_collection, timeseries_core_collection
        )

        mapped_chart = item.as_request_resource()
        mapped_chart.data.core_timeseries_collection = timeseries_core_collection
        mapped_chart.data.time_series_collection = None
        mapped_chart.data.source_collection = updated_source_collection
        return mapped_chart, issue

    def _create_timeseries_core_collection(
        self, time_series_collection: list[ChartTimeseries], issue: ChartMigrationIssue
    ) -> list[ChartCoreTimeseries]:
        timeseries_core_collection: list[ChartCoreTimeseries] = []
        for ts_item in time_series_collection or []:
            node_id, consumer_view_id = self._get_node_id_consumer_view_id(ts_item)

            if node_id is None:
                if ts_item.ts_id is not None:
                    issue.missing_timeseries_ids.append(ts_item.ts_id)
                elif ts_item.ts_external_id is not None:
                    issue.missing_timeseries_external_ids.append(ts_item.ts_external_id)
                else:
                    issue.missing_timeseries_identifier.append(ts_item.id or "unknown")
                continue

            core_timeseries = self._create_new_timeseries_core(ts_item, node_id, consumer_view_id)
            timeseries_core_collection.append(core_timeseries)
        return timeseries_core_collection

    def _create_new_timeseries_core(
        self, ts_item: ChartTimeseries, node_id: NodeReference, consumer_view_id: ViewReference | None
    ) -> ChartCoreTimeseries:
        dumped = ts_item.model_dump(mode="json", by_alias=True, exclude_unset=True)
        dumped["nodeReference"] = NodeId(space=node_id.space, external_id=node_id.external_id)
        dumped["viewReference"] = (
            ViewId(
                space=consumer_view_id.space, external_id=consumer_view_id.external_id, version=consumer_view_id.version
            )
            if consumer_view_id
            else None
        )
        new_uuid = str(uuid4())
        dumped["id"] = new_uuid
        dumped["type"] = "coreTimeseries"
        # We ignore extra here to only include the fields that are shared between ChartTimeseries and ChartCoreTimeseries
        core_timeseries = ChartCoreTimeseries.model_validate(dumped, extra="ignore")
        return core_timeseries

    def _get_node_id_consumer_view_id(
        self, ts_item: ChartTimeseries
    ) -> tuple[NodeReference | None, ViewReference | None]:
        """Look up the node ID and consumer view ID for a given timeseries item.

        Prioritizes lookup by internal ID, then by external ID.

        Args:
            ts_item: The ChartTimeseries item to look up.

        Returns:
            A tuple containing the consumer view ID and node ID, or None if not found.
        """
        node_id: NodeReference | None = None
        consumer_view_id: ViewReference | None = None
        for id_name, id_value in [("id", ts_item.ts_id), ("external_id", ts_item.ts_external_id)]:
            if id_value is None:
                continue
            arg = {id_name: id_value}
            node_id = self.client.migration.lookup.time_series(**arg)  # type: ignore[arg-type]
            consumer_view_id = self.client.migration.lookup.time_series.consumer_view(**arg)  # type: ignore[arg-type]
            if node_id is not None:
                break
        return node_id, consumer_view_id

    def _update_source_collection(
        self,
        source_collection: list[ChartSource],
        time_series_collection: list[ChartTimeseries],
        timeseries_core_collection: list[ChartCoreTimeseries],
    ) -> list[ChartSource]:
        remove_ids = {ts_item.id for ts_item in time_series_collection if ts_item.id is not None}
        updated_source_collection = [ts_item for ts_item in source_collection if ts_item.id not in remove_ids]
        for core_ts_item in timeseries_core_collection:
            # We cast there two as we set them in the _create_timeseries_core_collection method
            new_source_item = ChartSource(id=cast(str, core_ts_item.id), type=cast(str, core_ts_item.type))
            updated_source_collection.append(new_source_item)
        return updated_source_collection


class CanvasMapper(DataMapper[CanvasSelector, IndustrialCanvas, IndustrialCanvasApply]):
    # Note sequences are not supported in Canvas, so we do not include them here.
    asset_centric_resource_types = tuple(("asset", "event", "timeseries", "file"))
    DEFAULT_ASSET_VIEW = ViewReference(space="cdf_cdm", external_id="CogniteAsset", version="v1")
    DEFAULT_EVENT_VIEW = ViewReference(space="cdf_cdm", external_id="CogniteActivity", version="v1")
    DEFAULT_FILE_VIEW = ViewReference(space="cdf_cdm", external_id="CogniteFile", version="v1")
    DEFAULT_TIMESERIES_VIEW = ViewReference(space="cdf_cdm", external_id="CogniteTimeSeries", version="v1")

    def __init__(self, client: ToolkitClient, dry_run: bool, skip_on_missing_ref: bool = False) -> None:
        super().__init__(client)
        self.dry_run = dry_run
        self.skip_on_missing_ref = skip_on_missing_ref

    def map(self, source: Sequence[IndustrialCanvas]) -> Sequence[IndustrialCanvasApply | None]:
        self._populate_cache(source)
        output: list[IndustrialCanvasApply | None] = []
        issues: list[CanvasMigrationIssue] = []
        for item in source:
            mapped_item, issue = self._map_single_item(item)
            identifier = item.as_id()

            if issue.missing_reference_ids:
                self.logger.tracker.add_issue(identifier, "Missing reference IDs")

            if issue.has_issues:
                issues.append(issue)

            if mapped_item is None:
                self.logger.tracker.finalize_item(identifier, "failure")

            output.append(mapped_item)
        if issues:
            self.logger.log(issues)
        return output

    @property
    def lookup_methods(self) -> dict[str, Callable]:
        return {
            "asset": self.client.migration.lookup.assets,
            "event": self.client.migration.lookup.events,
            "timeseries": self.client.migration.lookup.time_series,
            "file": self.client.migration.lookup.files,
        }

    def _populate_cache(self, source: Sequence[IndustrialCanvas]) -> None:
        """Populate the internal cache with references from the source canvases.

        Note that the consumption views are also cached as part of the timeseries lookup.
        """
        ids_by_type: dict[str, set[int]] = defaultdict(set)
        for canvas in source:
            for ref in canvas.container_references or []:
                if ref.container_reference_type in self.asset_centric_resource_types:
                    ids_by_type[ref.container_reference_type].add(ref.resource_id)

        for resource_type, lookup_method in self.lookup_methods.items():
            ids = ids_by_type.get(resource_type)
            if ids:
                lookup_method(list(ids))

    def _get_node_id(self, resource_id: int, resource_type: str) -> NodeReference | None:
        """Look up the node ID for a given resource ID and type."""
        try:
            return self.lookup_methods[resource_type](resource_id)
        except KeyError:
            raise ToolkitValueError(f"Unsupported resource type '{resource_type}' for container reference migration.")

    def _get_consumer_view_id(self, resource_id: int, resource_type: str) -> ViewReference:
        """Look up the consumer view ID for a given resource ID and type."""
        lookup_map = {
            "asset": (self.client.migration.lookup.assets.consumer_view, self.DEFAULT_ASSET_VIEW),
            "event": (self.client.migration.lookup.events.consumer_view, self.DEFAULT_EVENT_VIEW),
            "timeseries": (self.client.migration.lookup.time_series.consumer_view, self.DEFAULT_TIMESERIES_VIEW),
            "file": (self.client.migration.lookup.files.consumer_view, self.DEFAULT_FILE_VIEW),
        }
        if lookup_tuple := lookup_map.get(resource_type):
            method, fallback = lookup_tuple
            return method(resource_id) or fallback

        raise ToolkitValueError(f"Unsupported resource type '{resource_type}' for container reference migration.")

    def _map_single_item(self, canvas: IndustrialCanvas) -> tuple[IndustrialCanvasApply | None, CanvasMigrationIssue]:
        update = canvas.as_write()
        issue = CanvasMigrationIssue(
            canvas_external_id=canvas.canvas.external_id, canvas_name=canvas.canvas.name, id=canvas.canvas.name
        )

        remaining_container_references: list[ContainerReferenceApply] = []
        new_fdm_references: list[FdmInstanceContainerReferenceApply] = []
        uuid_generator: dict[str, str] = defaultdict(lambda: str(uuid4()))
        for ref in update.container_references or []:
            if ref.container_reference_type not in self.asset_centric_resource_types:
                remaining_container_references.append(ref)
                continue
            node_id = self._get_node_id(ref.resource_id, ref.container_reference_type)
            if node_id is None:
                issue.missing_reference_ids.append(ref.as_asset_centric_id())
            else:
                consumer_view = self._get_consumer_view_id(ref.resource_id, ref.container_reference_type)
                fdm_ref = self.migrate_container_reference(
                    ref, canvas.canvas.external_id, node_id, consumer_view, uuid_generator
                )
                new_fdm_references.append(fdm_ref)
        if issue.missing_reference_ids and self.skip_on_missing_ref:
            return None, issue

        update.container_references = remaining_container_references
        update.fdm_instance_container_references.extend(new_fdm_references)
        if not self.dry_run:
            backup = canvas.as_write().create_backup()
            try:
                self.client.canvas.industrial.create(backup)
            except CogniteException as e:
                raise ToolkitMigrationError(
                    f"Failed to create backup for canvas '{canvas.canvas.name}': {e!s}. "
                ) from e

        # There might annotations or other components that reference the old IDs, so we need to update those as well.
        update = update.replace_ids(uuid_generator)

        return update, issue

    @classmethod
    def migrate_container_reference(
        cls,
        reference: ContainerReferenceApply,
        canvas_external_id: str,
        instance_id: NodeReference,
        consumer_view: ViewReference,
        uuid_generator: dict[str, str],
    ) -> FdmInstanceContainerReferenceApply:
        """Migrate a single container reference by replacing the asset-centric ID with the data model instance ID."""
        new_id = uuid_generator[reference.id_]
        new_external_id = f"{canvas_external_id}_{new_id}"
        return FdmInstanceContainerReferenceApply(
            external_id=new_external_id,
            id_=new_id,
            container_reference_type="fdmInstance",
            instance_space=instance_id.space,
            instance_external_id=instance_id.external_id,
            view_space=consumer_view.space,
            view_external_id=consumer_view.external_id,
            view_version=consumer_view.version,
            label=reference.label,
            properties_=reference.properties_,
            x=reference.x,
            y=reference.y,
            width=reference.width,
            height=reference.height,
            max_width=reference.max_width,
            max_height=reference.max_height,
        )


class ThreeDMapper(DataMapper[ThreeDSelector, ThreeDModelClassicResponse, ThreeDMigrationRequest]):
    def map(self, source: Sequence[ThreeDModelClassicResponse]) -> Sequence[ThreeDMigrationRequest | None]:
        self._populate_cache(source)
        output: list[ThreeDMigrationRequest | None] = []
        issues: list[ThreeDModelMigrationIssue] = []
        for item in source:
            mapped_item, issue = self._map_single_item(item)
            identifier = item.name

            if issue.error_message:
                for error in issue.error_message:
                    self.logger.tracker.add_issue(identifier, error)

            if issue.has_issues:
                issues.append(issue)

            if mapped_item is None:
                self.logger.tracker.finalize_item(identifier, "failure")
            output.append(mapped_item)
        if issues:
            self.logger.log(issues)
        return output

    def _populate_cache(self, source: Sequence[ThreeDModelClassicResponse]) -> None:
        dataset_ids: set[int] = set()
        for model in source:
            if model.data_set_id is not None:
                dataset_ids.add(model.data_set_id)
        self.client.migration.space_source.retrieve(list(dataset_ids))

    def _map_single_item(
        self, item: ThreeDModelClassicResponse
    ) -> tuple[ThreeDMigrationRequest | None, ThreeDModelMigrationIssue]:
        issue = ThreeDModelMigrationIssue(model_name=item.name, model_id=item.id, id=item.name)
        instance_space: str | None = None
        last_revision_id: int | None = None
        model_type: Literal["CAD", "PointCloud", "Image360"] | None = None
        if item.data_set_id is None:
            issue.error_message.append("3D model is not associated with any dataset.")
        else:
            space_source = self.client.migration.space_source.retrieve(item.data_set_id)
            if space_source is not None:
                instance_space = space_source.instance_space
        if instance_space is None and item.data_set_id is not None:
            issue.error_message.append(f"Missing instance space for dataset ID {item.data_set_id!r}")
        if item.last_revision_info is None:
            issue.error_message.append("3D model has no revisions.")
        else:
            model_type = self._get_type(item.last_revision_info)
            last_revision_id = item.last_revision_info.revision_id
            if last_revision_id is None:
                issue.error_message.append("3D model's last revision has no revision ID.")

        if model_type is None:
            issue.error_message.append("3D model's last revision has no recognized type.")

        if instance_space is None or last_revision_id is None or model_type is None or issue.has_issues:
            return None, issue

        mapped_request = ThreeDMigrationRequest(
            model_id=item.id,
            type=model_type,
            space=instance_space,
            revision=ThreeDRevisionMigrationRequest(
                space=instance_space,
                type=model_type,
                revision_id=last_revision_id,
                model=Model(
                    instance_id=InstanceIdentifier(
                        space=instance_space,
                        external_id=f"cog_3d_model_{item.id!s}",
                    )
                ),
            ),
        )
        return mapped_request, issue

    @staticmethod
    def _get_type(revision: RevisionStatus) -> Literal["CAD", "PointCloud", "Image360"] | None:
        types = revision.types or []
        if any("gltf-directory" in t for t in types):
            return "CAD"
        elif any("ept-pointcloud" in t for t in types):
            return "PointCloud"
        else:
            return None


class ThreeDAssetMapper(DataMapper[ThreeDSelector, AssetMappingClassicResponse, AssetMappingDMRequest]):
    def map(self, source: Sequence[AssetMappingClassicResponse]) -> Sequence[AssetMappingDMRequest | None]:
        output: list[AssetMappingDMRequest | None] = []
        issues: list[ThreeDModelMigrationIssue] = []
        self._populate_cache(source)
        for item in source:
            mapped_item, issue = self._map_single_item(item)
            identifier = f"AssetMapping_{item.model_id!s}_{item.revision_id!s}_{item.asset_id!s}"

            if issue.error_message:
                for error in issue.error_message:
                    self.logger.tracker.add_issue(identifier, error)

            if issue.has_issues:
                issues.append(issue)

            if mapped_item is None:
                self.logger.tracker.finalize_item(identifier, "failure")

            output.append(mapped_item)
        if issues:
            self.logger.log(issues)
        return output

    def _populate_cache(self, source: Sequence[AssetMappingClassicResponse]) -> None:
        asset_ids: set[int] = set()
        for mapping in source:
            if mapping.asset_id is not None:
                asset_ids.add(mapping.asset_id)
        self.client.migration.lookup.assets(list(asset_ids))

    def _map_single_item(
        self, item: AssetMappingClassicResponse
    ) -> tuple[AssetMappingDMRequest | None, ThreeDModelMigrationIssue]:
        issue = ThreeDModelMigrationIssue(
            model_name=f"AssetMapping_{item.model_id}", model_id=item.model_id, id=f"AssetMapping_{item.model_id}"
        )
        asset_instance_id = item.asset_instance_id
        if item.asset_id and asset_instance_id is None:
            asset_node_id = self.client.migration.lookup.assets(item.asset_id)
            if asset_node_id is None:
                issue.error_message.append(f"Missing asset instance for asset ID {item.asset_id!r}")
                return None, issue
            asset_instance_id = NodeReference(space=asset_node_id.space, external_id=asset_node_id.external_id)

        if asset_instance_id is None:
            issue.error_message.append("Neither assetInstanceId nor assetId provided for mapping.")
            return None, issue
        mapped_request = AssetMappingDMRequest(
            model_id=item.model_id,
            revision_id=item.revision_id,
            node_id=item.node_id,
            asset_instance_id=asset_instance_id,
        )
        return mapped_request, issue



from typing import Literal, TypeAlias

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource
from cognite_toolkit._cdf_tk.client.resource_classes.charts_data import ChartData
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

Visibility: TypeAlias = Literal["PUBLIC", "PRIVATE"]


class Chart(BaseModelObject, extra="allow"):
    external_id: str
    visibility: Visibility
    data: ChartData

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class ChartRequest(Chart, RequestResource, extra="allow"): ...


class ChartResponse(Chart, ResponseResource[ChartRequest], extra="allow"):
    created_time: int
    last_updated_time: int
    owner_id: str

    @classmethod
    def request_cls(cls) -> type[ChartRequest]:
        return ChartRequest



"""Annotations API for managing CDF annotations.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Annotations/operation/annotationsCreate
"""

from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import AnnotationFilter
from cognite_toolkit._cdf_tk.client.resource_classes.annotation import AnnotationRequest, AnnotationResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId


class AnnotationsAPI(CDFResourceAPI[AnnotationResponse]):
    """API for managing CDF annotations."""

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/annotations", item_limit=1000),
                "retrieve": Endpoint(method="POST", path="/annotations/byids", item_limit=1000),
                "update": Endpoint(method="POST", path="/annotations/update", item_limit=1000),
                "delete": Endpoint(method="POST", path="/annotations/delete", item_limit=1000),
                "list": Endpoint(method="POST", path="/annotations/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[AnnotationResponse]:
        return PagedResponse[AnnotationResponse].model_validate_json(response.body)

    def create(self, items: Sequence[AnnotationRequest]) -> list[AnnotationResponse]:
        """Create annotations in CDF.

        Args:
            items: List of AnnotationRequest objects to create.

        Returns:
            List of created AnnotationResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(self, items: Sequence[InternalId]) -> list[AnnotationResponse]:
        """Retrieve annotations from CDF by ID.

        Args:
            items: List of InternalId objects to retrieve.

        Returns:
            List of retrieved AnnotationResponse objects.
        """
        return self._request_item_response(items, method="retrieve")

    def update(
        self, items: Sequence[AnnotationRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[AnnotationResponse]:
        """Update annotations in CDF.

        Args:
            items: List of AnnotationRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated AnnotationResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalId]) -> None:
        """Delete annotations from CDF.

        Args:
            items: List of InternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: AnnotationFilter,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[AnnotationResponse]:
        """Get a page of annotations from CDF.

        Args:
            filter: AnnotationFilter to filter annotations.
            limit: Maximum number of annotations to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of AnnotationResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit, body={"filter": filter.dump()})

    def iterate(
        self,
        filter: AnnotationFilter,
        limit: int | None = None,
    ) -> Iterable[list[AnnotationResponse]]:
        """Iterate over all annotations in CDF.

        Args:
            filter: AnnotationFilter to filter annotations.
            limit: Maximum total number of annotations to return.

        Returns:
            Iterable of lists of AnnotationResponse objects.
        """
        return self._iterate(limit=limit, body={"filter": filter.dump()})

    def list(
        self,
        filter: AnnotationFilter,
        limit: int | None = None,
    ) -> list[AnnotationResponse]:
        """List all annotations in CDF.

        Args:
            filter: AnnotationFilter to filter annotations.
            limit: Maximum total number of annotations to return.

        Returns:
            List of AnnotationResponse objects.
        """
        return self._list(limit=limit, body={"filter": filter.dump()})



from typing import Generic, TypeVar

from pydantic import BaseModel, Field

T = TypeVar("T", bound=BaseModel)


class ResponseItems(BaseModel, Generic[T]):
    """A page of reference items from a paginated API response.

    Attributes:
        items: The list of reference items in this page.
    """

    items: list[T]


class PagedResponse(BaseModel, Generic[T]):
    items: list[T]
    next_cursor: str | None = Field(None, alias="nextCursor")



from typing import Any

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class BaseModelRequest(BaseModel):
    """Base class for all object. This includes resources and nested objects."""

    model_config = ConfigDict(alias_generator=to_camel, extra="ignore", populate_by_name=True)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the object to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.

        """
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True)



from abc import ABC
from typing import Annotated, Literal

from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import WorkflowVersionId as WorkflowVersionIdIdentifier

from .base import BaseModelResource, ToolkitResource


class WorkflowVersionId(BaseModelResource):
    workflow_external_id: str = Field(
        max_length=255,
        description="Identifier for a workflow. Must be unique for the project. No trailing or leading whitespace and no null characters allowed.",
    )
    version: str = Field(
        max_length=255,
        description="Identifier for a version. Must be unique for the workflow. No trailing or leading whitespace and no null characters allowed.",
    )


class CogniteFunctionRef(BaseModelResource):
    external_id: str = Field(
        description="The external id of the Cognite Function in the project. This can be either a function external ID or a reference like ${myTaskExternalId.output.someKey}"
    )
    data: str | JsonValue | None = Field(
        None, description="Input data that will be passed to the Cognite Function. Limited to 100KB in size."
    )


class FunctionTaskParameters(BaseModelResource):
    function: CogniteFunctionRef
    is_async_complete: bool = Field(
        False, description="Defines if the execution of the task should be completed asynchronously."
    )


class TransformationRef(BaseModelResource):
    external_id: str = Field(
        description="The external id of the Transformation in the project. This can be either a transformation external ID or a reference like ${myTaskExternalId.output.someKey}"
    )
    concurrency_policy: Literal["fail", "waitForCurrent", "restartAfterCurrent"] = Field(
        "fail",
        description="""Determines the behavior of the task if the Transformation is already running.

fail: The task fails if another instance of the Transformation is currently running.
waitForCurrent: The task will pause and wait for the already running Transformation to complete. Once completed, the task is completed. This mode is useful for preventing redundant Transformation runs.
restartAfterCurrent: The task waits for the ongoing Transformation to finish. After completion, the task restarts the Transformation. This mode ensures that the most recent data can be used by following tasks.""",
    )
    use_transformation_credentials: bool = Field(
        False,
        description="If set to true, the transformation will run using the client credentials configured on the transformation. If set to false, the transformation will run using the client credentials used to trigger the workflow.",
    )


class TransformationTaskParameters(BaseModelResource):
    transformation: TransformationRef


class CDFRequest(BaseModelResource):
    resource_path: str = Field(
        description="The path of the request. The path should be prefixed by {cluster}.cognitedata.com/api/v1/project/{project} based on the relevant cluster and project. It can also contain references like ${myTaskExternalId.output.someKey}"
    )
    method: Literal["POST", "GET", "PUT"] | str = Field(
        description="The HTTP method of the request. It can also be a reference like ${myTaskExternalId.output.someKey}"
    )
    query_parameters: dict[str, JsonValue] | str | None = Field(
        None,
        description="The query parameters of the request. It can also be a reference like ${myTaskExternalId.output.someKey}",
    )
    body: JsonValue | str | None = Field(
        None, description="The body of the request. It can also be a reference like ${myTaskExternalId.output.someKey}"
    )
    request_timeout_in_millis: float | str | None = Field(
        None,
        description="The timeout for the request in milliseconds. It can also be a reference like ${myTaskExternalId.output.someKey}",
    )
    cdf_version_header: Literal["alpha", "beta"] | str | None = Field(
        None, description="The Cognite Data Fusion version header to use for the request."
    )


class CDFTaskParameters(BaseModelResource):
    cdf_request: CDFRequest


class DynamicRef(BaseModelResource):
    tasks: str = Field(
        description="A Reference is an expression that allows dynamically injecting input to a task during execution. References can be used to reference the input of the Workflow, the output of a previous task in the Workflow, or the input of a previous task in the Workflow. Note that the injected value must be valid in the context of the property it is injected into. Example Task reference: ${myTaskExternalId.output.someKey} Example Workflow input reference: ${workflow.input.myKey}"
    )


class DynamicTaskParameters(BaseModelResource):
    dynamic: DynamicRef = Field(description="Reference to another task to use as the definition for this task.")


class SubworkflowInlineTasks(BaseModelResource):
    """Inline definition of tasks for a subworkflow."""

    tasks: "list[Task]" = Field(description="Inline definition of tasks for the subworkflow.")


# SubworkflowRef can be either a reference to an existing workflow version OR an inline definition of tasks
SubworkflowRef = Annotated[
    WorkflowVersionId | SubworkflowInlineTasks,
    Field(
        description="Reference to the subworkflow to execute. This can be either a reference to an existing workflow version (with workflowExternalId and version) or an inline definition of tasks."
    ),
]


class SubworkflowTaskParameters(BaseModelResource):
    subworkflow: SubworkflowRef = Field(description="Reference to the subworkflow to execute.")


class SimulatorInputUnit(BaseModelResource):
    name: str = Field(description="Name of the unit.")


class SimulatorInput(BaseModelResource):
    reference_id: str = Field(description="Reference id of the value to override.")
    value: str | int | float | list[str] | list[int] | list[float] = Field(
        description="Override the value used for a simulation run."
    )
    unit: SimulatorInputUnit | None = Field(None, description="Override the unit of the value")


class SimulationRef(BaseModelResource):
    routine_external_id: str = Field(description="The external id of the routine to be executed.")
    run_time: int | None = Field(
        None,
        description="Run time in milliseconds. Reference timestamp used for data pre-processing and data sampling.",
    )
    inputs: list[SimulatorInput] | None = Field(
        None, description="List of inputs to be provided to the simulation.", max_length=200
    )


class SimulationTaskParameters(BaseModelResource):
    simulation: SimulationRef = Field(description="Reference to the simulation to execute.")


class TaskId(BaseModelResource):
    external_id: str = Field(
        max_length=255, description="The external ID provided by the client. Must be unique for the resource type."
    )


class TaskDefinition(BaseModelResource, ABC):
    external_id: str = Field(
        max_length=255,
        description="Identifier for the task. Must be unique within the version. No trailing or leading whitespace and no null characters allowed.",
    )
    type: str
    name: str | None = Field(
        default=None,
        max_length=255,
        description="Readable name meant for use in UIs",
    )
    description: str | None = Field(
        default=None,
        max_length=500,
        description="Description of the intention of the task",
    )
    retries: int = Field(
        3,
        ge=0,
        le=10,
        description="Number of times to retry the task if it fails. If set to 0, the task will not be retried. The behavior for timeouts and retries is defined by the onFailure parameter, refer to it for more information.",
    )
    timeout: int = Field(
        3600,
        ge=60,
        le=86400,
        description="TTimeout in seconds. After this time, the task will be marked as TIMED_OUT. By default, the task won't be retried upon timeout. Use the onFailure parameter to change this behavior.",
    )
    on_failure: Literal["abortWorkflow", "skipTask"] | None = Field(
        None,
        description="Defines the behavior when a task fails. If set to abortWorkflow, the entire workflow will be marked as FAILED when the task fails. If set to skipTask, the task will be marked as FAILED, but the workflow will continue executing subsequent tasks that are not dependent on the failed task.",
    )
    depends_on: list[TaskId] | None = Field(
        None,
        description="The tasks that must be completed before this task can be executed.",
        max_length=100,
    )


class FunctionTask(TaskDefinition):
    type: Literal["function"] = "function"
    parameters: FunctionTaskParameters


class TransformationTask(TaskDefinition):
    type: Literal["transformation"] = "transformation"
    parameters: TransformationTaskParameters


class CDFTask(TaskDefinition):
    type: Literal["cdfRequest"] = "cdfRequest"
    parameters: CDFTaskParameters


class DynamicTask(TaskDefinition):
    type: Literal["dynamic"] = "dynamic"
    parameters: DynamicTaskParameters


class SubworkflowTask(TaskDefinition):
    type: Literal["subworkflow"] = "subworkflow"
    parameters: SubworkflowTaskParameters


class SimulationTask(TaskDefinition):
    type: Literal["simulation"] = "simulation"
    parameters: SimulationTaskParameters


Task = Annotated[
    FunctionTask | TransformationTask | CDFTask | DynamicTask | SubworkflowTask | SimulationTask,
    Field(discriminator="type"),
]


class WorkflowDefinition(BaseModelResource):
    description: str | None = Field(
        default=None,
        max_length=500,
        description="The description of the workflow version.",
    )
    tasks: list[Task]


class WorkflowVersionYAML(ToolkitResource):
    workflow_external_id: str = Field(
        max_length=255,
        description="Identifier for a workflow. Must be unique for the project. No trailing or leading whitespace and no null characters allowed.",
    )
    version: str = Field(
        max_length=255,
        description="Identifier for a version. Must be unique for the workflow. No trailing or leading whitespace and no null characters allowed.",
    )
    workflow_definition: WorkflowDefinition

    def as_id(self) -> WorkflowVersionIdIdentifier:
        return WorkflowVersionIdIdentifier(workflow_external_id=self.workflow_external_id, version=self.version)



from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalIdUnwrapped
from cognite_toolkit._cdf_tk.client.resource_classes.securitycategory import (
    SecurityCategoryRequest,
    SecurityCategoryResponse,
)


class SecurityCategoriesAPI(CDFResourceAPI[SecurityCategoryResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/securitycategories", item_limit=1000),
                "delete": Endpoint(method="POST", path="/securitycategories/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/securitycategories", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[SecurityCategoryResponse]:
        return PagedResponse[SecurityCategoryResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalIdUnwrapped]:
        return ResponseItems[InternalIdUnwrapped].model_validate_json(response.body)

    def create(self, items: Sequence[SecurityCategoryRequest]) -> list[SecurityCategoryResponse]:
        """Create security categories in CDF.

        Args:
            items: List of SecurityCategoryRequest objects to create.
        Returns:
            List of created SecurityCategoryResponse objects.
        """
        return self._request_item_response(items, "create")

    def delete(self, items: Sequence[InternalIdUnwrapped]) -> None:
        """Delete security categories from CDF.

        Args:
            items: List of InternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        sort: Literal["ASC", "DESC"] = "ASC",
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[SecurityCategoryResponse]:
        """Iterate over all security categories in CDF.

        Args:
            sort: Sort descending or ascending.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of SecurityCategoryResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit, params={"sort": sort})

    def iterate(
        self,
        sort: Literal["ASC", "DESC"] = "ASC",
        limit: int | None = 100,
    ) -> Iterable[list[SecurityCategoryResponse]]:
        """Iterate over all security categories in CDF.

        Args:
            sort: Sort descending or ascending.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of SecurityCategoryResponse objects.
        """
        return self._iterate(limit=limit, params={"sort": sort})

    def list(
        self,
        limit: int | None = 100,
    ) -> list[SecurityCategoryResponse]:
        """List all security categories in CDF.

        Returns:
            List of SecurityCategoryResponse objects.
        """
        return self._list(limit=limit)



import sys
from pathlib import Path

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from pydantic import BaseModel, ConfigDict

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.constants import DEFAULT_ENV
from cognite_toolkit._cdf_tk.data_classes import (
    BuildConfigYAML,
)
from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning, WarningList
from cognite_toolkit._cdf_tk.utils.modules import parse_user_selected_modules


class BuildParameters(BaseModel):
    """Input to the build process."""

    # need this until we turn BuildConfigYaml and ToolkitClient into Pydantic models
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    organization_dir: Path
    build_dir: Path
    build_env_name: str
    config: BuildConfigYAML
    client: ToolkitClient | None = None
    warnings: WarningList[ToolkitWarning] | None = None
    user_selected: list[str | Path] | None = None

    @classmethod
    def load(
        cls,
        organization_dir: Path,
        build_dir: Path,
        build_env_name: str | None,
        client: ToolkitClient | None,
        user_selected: list[str | Path] | None = None,
    ) -> Self:
        resolved_org_dir = Path.cwd() if organization_dir in {Path("."), Path("./")} else organization_dir
        resolved_env = build_env_name or DEFAULT_ENV
        config, warnings = cls._load_config(resolved_org_dir, resolved_env, user_selected)
        return cls(
            organization_dir=resolved_org_dir,
            build_dir=build_dir,
            build_env_name=resolved_env,
            config=config,
            client=client,
            warnings=warnings,
            user_selected=user_selected,
        )

    @classmethod
    def _load_config(
        cls, organization_dir: Path, build_env_name: str, user_selected: list[str | Path] | None
    ) -> tuple[BuildConfigYAML, WarningList[ToolkitWarning]]:
        warnings: WarningList[ToolkitWarning] = WarningList[ToolkitWarning]()
        if (organization_dir / BuildConfigYAML.get_filename(build_env_name or DEFAULT_ENV)).exists():
            config = BuildConfigYAML.load_from_directory(organization_dir, build_env_name or DEFAULT_ENV)
        else:
            # Loads the default environment
            config = BuildConfigYAML.load_default(organization_dir)
        if user_selected:
            config.environment.selected = list(set(parse_user_selected_modules(list(user_selected), organization_dir)))
        if environment_warning := config.validate_environment():
            warnings.append(environment_warning)
        return config, warnings



import json
from collections.abc import Hashable, Iterable, Sequence
from typing import Any

from cognite.client.data_classes import capabilities
from cognite.client.data_classes.capabilities import Capability

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.robotics import (
    RobotCapabilityRequest,
    RobotCapabilityResponse,
    RobotDataPostProcessingRequest,
    RobotDataPostProcessingResponse,
    RobotFrameRequest,
    RobotFrameResponse,
    RobotLocationRequest,
    RobotLocationResponse,
    RobotMapRequest,
    RobotMapResponse,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import (
    RobotCapabilityYAML,
    RobotDataPostProcessingYAML,
    RobotFrameYAML,
    RobotLocationYAML,
    RobotMapYAML,
)
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable


class RoboticFrameCRUD(ResourceCRUD[ExternalId, RobotFrameRequest, RobotFrameResponse]):
    folder_name = "robotics"
    resource_cls = RobotFrameResponse
    resource_write_cls = RobotFrameRequest
    kind = "Frame"
    yaml_cls = RobotFrameYAML
    _doc_url = "Frames/operation/createFrames"

    @property
    def display_name(self) -> str:
        return "robotics frames"

    @classmethod
    def get_id(cls, item: RobotFrameRequest | RobotFrameResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RobotFrameRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        return capabilities.RoboticsAcl(
            [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
            ],
            capabilities.RoboticsAcl.Scope.All(),
        )

    def dump_resource(self, resource: RobotFrameResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if dumped.get("transform") is None and "transform" not in (local or {}):
            # Default value set on the server side.
            dumped.pop("transform", None)
        return dumped

    def create(self, items: Sequence[RobotFrameRequest]) -> list[RobotFrameResponse]:
        return self.client.tool.robotics.frames.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[RobotFrameResponse]:
        return self.client.tool.robotics.frames.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[RobotFrameRequest]) -> list[RobotFrameResponse]:
        return self.client.tool.robotics.frames.update(items, mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.robotics.frames.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RobotFrameResponse]:
        for frames in self.client.tool.robotics.frames.iterate(limit=None):
            yield from frames


class RoboticLocationCRUD(ResourceCRUD[ExternalId, RobotLocationRequest, RobotLocationResponse]):
    folder_name = "robotics"
    resource_cls = RobotLocationResponse
    resource_write_cls = RobotLocationRequest
    kind = "Location"
    yaml_cls = RobotLocationYAML
    _doc_url = "Locations/operation/createLocations"

    @property
    def display_name(self) -> str:
        return "robotics locations"

    @classmethod
    def get_id(cls, item: RobotLocationRequest | RobotLocationResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RobotLocationRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def create(self, items: Sequence[RobotLocationRequest]) -> list[RobotLocationResponse]:
        return self.client.tool.robotics.locations.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[RobotLocationResponse]:
        return self.client.tool.robotics.locations.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[RobotLocationRequest]) -> list[RobotLocationResponse]:
        return self.client.tool.robotics.locations.update(items, mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.robotics.locations.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RobotLocationResponse]:
        for locations in self.client.tool.robotics.locations.iterate(limit=None):
            yield from locations

    def dump_resource(self, resource: RobotLocationResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        return resource.as_request_resource().dump()

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> RobotLocationRequest:
        return RobotLocationRequest.model_validate(resource)


class RoboticsDataPostProcessingCRUD(
    ResourceCRUD[ExternalId, RobotDataPostProcessingRequest, RobotDataPostProcessingResponse]
):
    folder_name = "robotics"
    resource_cls = RobotDataPostProcessingResponse
    resource_write_cls = RobotDataPostProcessingRequest
    kind = "DataPostProcessing"
    yaml_cls = RobotDataPostProcessingYAML
    _doc_url = "DataPostProcessing/operation/createDataPostProcessing"

    @property
    def display_name(self) -> str:
        return "robotics data postprocessing"

    @classmethod
    def get_id(cls, item: RobotDataPostProcessingRequest | RobotDataPostProcessingResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RobotDataPostProcessingRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def create(self, items: Sequence[RobotDataPostProcessingRequest]) -> list[RobotDataPostProcessingResponse]:
        return self.client.tool.robotics.data_postprocessing.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[RobotDataPostProcessingResponse]:
        return self.client.tool.robotics.data_postprocessing.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[RobotDataPostProcessingRequest]) -> list[RobotDataPostProcessingResponse]:
        # There is a bug in the /update endpoint that requires the input_schema to be a string
        # and not an object. This is a workaround until the bug is fixed.
        # We do the serialization to avoid modifying the original object.
        to_update = []
        for item in items:
            if isinstance(item.input_schema, dict):
                update = RobotDataPostProcessingRequest.model_validate(item.dump())
                update.input_schema = json.dumps(item.input_schema)  # type: ignore[assignment]
                to_update.append(update)
            else:
                to_update.append(item)

        return self.client.tool.robotics.data_postprocessing.update(to_update, mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.robotics.data_postprocessing.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RobotDataPostProcessingResponse]:
        for items in self.client.tool.robotics.data_postprocessing.iterate(limit=None):
            yield from items

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "inputSchema":
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


class RobotCapabilityCRUD(ResourceCRUD[ExternalId, RobotCapabilityRequest, RobotCapabilityResponse]):
    folder_name = "robotics"
    resource_cls = RobotCapabilityResponse
    resource_write_cls = RobotCapabilityRequest
    kind = "RobotCapability"
    yaml_cls = RobotCapabilityYAML
    _doc_url = "RobotCapabilities/operation/createRobotCapabilities"

    @property
    def display_name(self) -> str:
        return "robotics robot capabilities"

    @classmethod
    def get_id(cls, item: RobotCapabilityRequest | RobotCapabilityResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RobotCapabilityRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def create(self, items: Sequence[RobotCapabilityRequest]) -> list[RobotCapabilityResponse]:
        return self.client.tool.robotics.capabilities.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[RobotCapabilityResponse]:
        return self.client.tool.robotics.capabilities.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[RobotCapabilityRequest]) -> list[RobotCapabilityResponse]:
        # There is a bug in the /update endpoint that requires the input_schema to be a string
        # and not an object. This is a workaround until the bug is fixed.
        # We do the serialization to avoid modifying the original object.
        to_update = []
        for item in items:
            if isinstance(item.input_schema, dict) or isinstance(item.data_handling_schema, dict):
                update = RobotCapabilityRequest.model_validate(item.dump())
                if isinstance(item.data_handling_schema, dict):
                    update.data_handling_schema = json.dumps(item.data_handling_schema)  # type: ignore[assignment]
                if isinstance(item.input_schema, dict):
                    update.input_schema = json.dumps(item.input_schema)  # type: ignore[assignment]
                to_update.append(update)
            else:
                to_update.append(item)

        return self.client.tool.robotics.capabilities.update(to_update, mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.robotics.capabilities.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RobotCapabilityResponse]:
        for items in self.client.tool.robotics.capabilities.iterate(limit=None):
            yield from items

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] in {"inputSchema", "dataHandlingSchema"}:
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


class RoboticMapCRUD(ResourceCRUD[ExternalId, RobotMapRequest, RobotMapResponse]):
    folder_name = "robotics"
    resource_cls = RobotMapResponse
    resource_write_cls = RobotMapRequest
    kind = "Map"
    dependencies = frozenset({RoboticFrameCRUD, RoboticLocationCRUD})
    yaml_cls = RobotMapYAML
    _doc_url = "Maps/operation/createMaps"

    @property
    def display_name(self) -> str:
        return "robotics maps"

    @classmethod
    def get_id(cls, item: RobotMapRequest | RobotMapResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RobotMapRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def dump_resource(self, resource: RobotMapResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dump = resource.as_request_resource().dump()
        local = local or {}
        if dump.get("scale") == 1.0 and "scale" not in local:
            # Default value set on the server side.
            del dump["scale"]
        for key in [
            "data",
            "frameExternalId",
            "locationExternalId",
        ]:
            if dump.get(key) is None and key not in (local or {}):
                # Key set to null on the server side.
                dump.pop(key, None)
        return dump

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> RobotMapRequest:
        return RobotMapRequest.model_validate(resource)

    def create(self, items: Sequence[RobotMapRequest]) -> list[RobotMapResponse]:
        return self.client.tool.robotics.maps.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[RobotMapResponse]:
        return self.client.tool.robotics.maps.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[RobotMapRequest]) -> list[RobotMapResponse]:
        return self.client.tool.robotics.maps.update(items, mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.robotics.maps.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RobotMapResponse]:
        for maps in self.client.tool.robotics.maps.iterate(limit=None):
            yield from maps



"""Relationships API for managing CDF relationships.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Relationships/operation/createRelationships
"""

from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.relationship import RelationshipRequest, RelationshipResponse


class RelationshipsAPI(CDFResourceAPI[RelationshipResponse]):
    """API for managing CDF relationships."""

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/relationships", item_limit=1000),
                "retrieve": Endpoint(method="POST", path="/relationships/byids", item_limit=1000),
                "update": Endpoint(method="POST", path="/relationships/update", item_limit=1000),
                "delete": Endpoint(method="POST", path="/relationships/delete", item_limit=1000),
                "list": Endpoint(method="POST", path="/relationships/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[RelationshipResponse]:
        return PagedResponse[RelationshipResponse].model_validate_json(response.body)

    def create(self, items: Sequence[RelationshipRequest]) -> list[RelationshipResponse]:
        """Create relationships in CDF.

        Args:
            items: List of RelationshipRequest objects to create.

        Returns:
            List of created RelationshipResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False, fetch_resources: bool = False
    ) -> list[RelationshipResponse]:
        """Retrieve relationships from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
            fetch_resources: If true, will try to fetch the resources referred to in the relationship,
                based on the users access rights. Will silently fail to attach the resources
                if the user lacks access to some of them.

        Returns:
            List of retrieved RelationshipResponse objects.
        """
        return self._request_item_response(
            items,
            method="retrieve",
            extra_body={"ignoreUnknownIds": ignore_unknown_ids, "fetchResources": fetch_resources},
        )

    def update(
        self, items: Sequence[RelationshipRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[RelationshipResponse]:
        """Update relationships in CDF.

        Args:
            items: List of RelationshipRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated RelationshipResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete relationships from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        filter: ClassicFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RelationshipResponse]:
        """Get a page of relationships from CDF.

        Args:
            limit: Maximum number of relationships to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RelationshipResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter.dump() if filter else None},
        )

    def iterate(
        self,
        filter: ClassicFilter | None = None,
        limit: int | None = None,
    ) -> Iterable[list[RelationshipResponse]]:
        """Iterate over all relationships in CDF.

        Args:
            limit: Maximum total number of relationships to return.

        Returns:
            Iterable of lists of RelationshipResponse objects.
        """
        return self._iterate(limit=limit, body={"filter": filter.dump() if filter else None})

    def list(self, filter: ClassicFilter | None = None, limit: int | None = None) -> list[RelationshipResponse]:
        """List all relationships in CDF.

        Args:
            limit: Maximum total number of relationships to return.

        Returns:
            List of RelationshipResponse objects.
        """
        return self._list(limit=limit, body={"filter": filter.dump() if filter else None})



import builtins
from typing import Any, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import ResponseResource, UpdatableRequestResource

from .identifiers import ExternalId


class SimulatorModelRequest(UpdatableRequestResource):
    # The 'id' field is not part of the request when creating a new resource,
    # but is needed when updating an existing resource.
    id: int | None = Field(default=None, exclude=True)
    external_id: str
    simulator_external_id: str
    name: str
    description: str | None = None
    data_set_id: int
    type: str

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        if self.id is None:
            raise ValueError("id must be provided to create an update dictionary")
        return {
            "id": self.id,
            "update": {
                "name": {"set": self.name},
                **{"description": {"set": self.description} if self.description is not None else {}},
            },
        }


class SimulatorModelResponse(ResponseResource[SimulatorModelRequest]):
    id: int
    external_id: str
    simulator_external_id: str
    name: str
    description: str | None = None
    data_set_id: int
    type: str | None = None
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> builtins.type[SimulatorModelRequest]:
        return SimulatorModelRequest

    def as_request_resource(self) -> SimulatorModelRequest:
        if self.type is None:
            raise ValueError("Cannot convert SimulatorModelResponse to SimulatorModelRequest when type is None")
        return SimulatorModelRequest.model_validate(self.dump(), extra="ignore")



from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class Credentials(BaseModelObject):
    nonce: str


class HostedExtractorDestination(BaseModelObject):
    external_id: str
    target_data_set_id: int | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class HostedExtractorDestinationRequest(HostedExtractorDestination, UpdatableRequestResource):
    credentials: Credentials | None = None


class HostedExtractorDestinationResponse(
    HostedExtractorDestination, ResponseResource[HostedExtractorDestinationRequest]
):
    session_id: int | None = None
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[HostedExtractorDestinationRequest]:
        return HostedExtractorDestinationRequest



from ._build import BuildFolder, BuildParameters, BuildSourceFiles, BuiltModule
from ._config import ConfigYAML
from ._insights import ConsistencyError, Insight, InsightList, ModelSyntaxError, Recommendation
from ._module import (
    BuildVariable,
    FailedReadResource,
    Module,
    ModuleSource,
    ReadResource,
    ResourceType,
    SuccessfulReadResource,
)
from ._types import AbsoluteDirPath, RelativeDirPath, RelativeFilePath, ValidationType

__all__ = [
    "AbsoluteDirPath",
    "BuildFolder",
    "BuildParameters",
    "BuildSourceFiles",
    "BuildVariable",
    "BuiltModule",
    "ConfigYAML",
    "ConsistencyError",
    "FailedReadResource",
    "Insight",
    "InsightList",
    "ModelSyntaxError",
    "Module",
    "ModuleSource",
    "ReadResource",
    "Recommendation",
    "RelativeDirPath",
    "RelativeFilePath",
    "ResourceType",
    "SuccessfulReadResource",
    "ValidationType",
]



from typing import Annotated, Any, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.constants import DM_EXTERNAL_ID_PATTERN, DM_VERSION_PATTERN, SPACE_FORMAT_PATTERN

from .base import BaseModelResource, ToolkitResource


class AgentToolModelResource(BaseModelResource, extra="allow"): ...


# --- Agent tool definitions ---


class AgentToolDefinition(AgentToolModelResource):
    type: str
    name: str = Field(
        description="A name for the tool, unique within the agent.",
        min_length=1,
        max_length=64,
        pattern=r"^[^\x00]{1,64}$",
    )
    description: str = Field(
        description="A description of how the tool helps the language model understand when and how to use the tool.",
        min_length=10,
        max_length=1024,
    )


class AnalyzeImage(AgentToolDefinition):
    type: Literal["analyzeImage"] = "analyzeImage"


class AnalyzeTimeSeries(AgentToolDefinition):
    type: Literal["analyzeTimeSeries"] = "analyzeTimeSeries"


class AskDocument(AgentToolDefinition):
    type: Literal["askDocument"] = "askDocument"


class CallFunctionConfig(AgentToolModelResource):
    external_id: str = Field(
        description="The external id of an existing Cognite Function in your CDF project.",
        min_length=1,
        max_length=255,
    )
    max_polling_time: int = Field(
        default=540,
        description="The maximum time in seconds to poll for the Cognite Function to complete.",
        gt=0,
        lt=541,
    )
    schema_: dict[str, Any] = Field(
        alias="schema",
        description="The Cognite Function's params specified as a JSON schema.",
    )


class CallFunction(AgentToolDefinition):
    type: Literal["callFunction"] = "callFunction"
    configuration: CallFunctionConfig = Field(
        description="Configuration for the Call Function tool.",
    )


class CallRestApi(AgentToolDefinition):
    type: Literal["callRestApi"] = "callRestApi"


class ExamineDataSemantically(AgentToolDefinition):
    type: Literal["examineDataSemantically"] = "examineDataSemantically"


class AgentDataModel(AgentToolModelResource):
    space: str = Field(
        description="The space the data model is in.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="The external ID of the data model.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="The version of the data model.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    view_external_ids: list[str] = Field(description="The views of the data model.", min_length=1, max_length=10)


class AgentInstanceSpacesDefinition(AgentToolModelResource):
    type: str


class AllInstanceSpaces(AgentInstanceSpacesDefinition):
    type: Literal["all"] = "all"


class ManualInstanceSpaces(AgentInstanceSpacesDefinition):
    type: Literal["manual"] = "manual"
    spaces: list[str]


AgentInstanceSpaces = Annotated[
    AllInstanceSpaces | ManualInstanceSpaces,
    Field(discriminator="type"),
]


class QueryKnowledgeGraphConfig(AgentToolModelResource):
    data_models: list[AgentDataModel] = Field(
        description="List of relevant data models.",
        min_length=1,
        max_length=80,
    )
    instance_spaces: AgentInstanceSpaces
    version: Literal["v1", "v2"] = Field(
        "v2",
        description="The version of the query generation strategy to use. "
        "A higher number does not necessarily mean a better query.",
    )


class QueryKnowledgeGraph(AgentToolDefinition):
    type: Literal["queryKnowledgeGraph"] = "queryKnowledgeGraph"
    configuration: QueryKnowledgeGraphConfig = Field(
        description="Configuration for the Query Knowledge Graph tool.",
    )


class QueryTimeSeriesDatapoints(AgentToolDefinition):
    type: Literal["queryTimeSeriesDatapoints"] = "queryTimeSeriesDatapoints"


class RunPythonCode(AgentToolDefinition):
    type: Literal["runPythonCode"] = "runPythonCode"


class SummarizeDocument(AgentToolDefinition):
    type: Literal["summarizeDocument"] = "summarizeDocument"


class TimeSeriesAnalysis(AgentToolDefinition):
    type: Literal["timeSeriesAnalysis"] = "timeSeriesAnalysis"


KNOWN_TOOLS: frozenset[str] = frozenset(
    {
        "analyzeImage",
        "analyzeTimeSeries",
        "askDocument",
        "callFunction",
        "callRestApi",
        "examineDataSemantically",
        "queryKnowledgeGraph",
        "queryTimeSeriesDatapoints",
        "runPythonCode",
        "summarizeDocument",
        "timeSeriesAnalysis",
    }
)


AgentTool = Annotated[
    AnalyzeImage
    | AnalyzeTimeSeries
    | AskDocument
    | CallFunction
    | CallRestApi
    | ExamineDataSemantically
    | QueryKnowledgeGraph
    | QueryTimeSeriesDatapoints
    | RunPythonCode
    | SummarizeDocument
    | TimeSeriesAnalysis,
    Field(discriminator="type"),
]


# --- Agent YAML resource ---


Model = Literal[
    "azure/o3",
    "azure/o4-mini",
    "azure/gpt-4o",
    "azure/gpt-4o-mini",
    "azure/gpt-4.1",
    "azure/gpt-4.1-nano",
    "azure/gpt-4.1-mini",
    "azure/gpt-5",
    "azure/gpt-5-mini",
    "azure/gpt-5-nano",
    "azure/gpt-5.1",
    "gcp/claude-4.5-sonnet",
    "gcp/claude-4.5-haiku",
    "gcp/gemini-2.5-pro",
    "gcp/gemini-2.5-flash",
    "aws/claude-4.5-sonnet",
    "aws/claude-4.5-haiku",
    "aws/claude-4-sonnet",
    "aws/claude-3.5-sonnet",
]


class AgentYAML(ToolkitResource):
    """Atlas AI Agent"""

    external_id: str = Field(
        description="An external ID that uniquely identifies the agent.",
        min_length=1,
        max_length=128,
        pattern=r"^[^\x00]{1,128}$",
    )
    name: str = Field(
        description="A descriptive name intended for use in user interfaces.",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="A human-readable description of what the agent does, used for documentation only. "
        "This description is not used by the language model.",
        max_length=1024,
    )
    instructions: str | None = Field(
        default=None,
        description="The instructions for the agent prompt the language model to understand "
        "the agent's goals and how to achieve them.",
        max_length=32000,
    )
    model: Model = Field(
        "azure/gpt-4o-mini", description="The name of the model to use. Defaults to your CDF project's default model."
    )
    tools: list[AgentTool] | None = Field(None, description="A list of tools available to the agent.", max_length=20)
    runtime_version: str | None = Field(None, description="The runtime version")

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



from enum import Enum
from pathlib import Path
from typing import Annotated, Any

import questionary
import typer
from questionary import Choice
from rich import print

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import RawTableId
from cognite_toolkit._cdf_tk.commands import DownloadCommand
from cognite_toolkit._cdf_tk.constants import DATA_DEFAULT_DIR
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.storageio import (
    AssetIO,
    CanvasIO,
    ChartIO,
    DatapointsIO,
    DataSelector,
    EventIO,
    FileContentIO,
    FileMetadataIO,
    HierarchyIO,
    InstanceIO,
    RawIO,
    StorageIO,
    TimeSeriesIO,
)
from cognite_toolkit._cdf_tk.storageio.selectors import (
    AssetSubtreeSelector,
    CanvasExternalIdSelector,
    CanvasSelector,
    ChartExternalIdSelector,
    ChartSelector,
    DataPointsDataSetSelector,
    DataSetSelector,
    FileIdentifierSelector,
    InstanceSelector,
    InstanceSpaceSelector,
    InstanceViewSelector,
    RawTableSelector,
    SelectedTable,
    SelectedView,
)
from cognite_toolkit._cdf_tk.storageio.selectors._file_content import FileInternalID
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetCentricInteractiveSelect,
    AssetInteractiveSelect,
    DataModelingSelect,
    EventInteractiveSelect,
    FileMetadataInteractiveSelect,
    InteractiveCanvasSelect,
    InteractiveChartSelect,
    RawTableInteractiveSelect,
    TimeSeriesInteractiveSelect,
    ViewSelectFilter,
)


class RawFormats(str, Enum):
    ndjson = "ndjson"
    yaml = "yaml"


class AssetCentricFormats(str, Enum):
    csv = "csv"
    parquet = "parquet"
    ndjson = "ndjson"


class FileContentFormats(str, Enum):
    ndjson = "ndjson"


class HierarchyFormats(str, Enum):
    ndjson = "ndjson"


class DatapointFormats(str, Enum):
    csv = "csv"
    parquet = "parquet"


class DatapointsDataTypes(str, Enum):
    numeric = "numeric"
    string = "string"


class InstanceFormats(str, Enum):
    ndjson = "ndjson"


class ChartFormats(str, Enum):
    ndjson = "ndjson"


class CanvasFormats(str, Enum):
    ndjson = "ndjson"


class InstanceTypes(str, Enum):
    node = "node"
    edge = "edge"


class CompressionFormat(str, Enum):
    gzip = "gzip"
    none = "none"


DEFAULT_DOWNLOAD_DIR = Path(DATA_DEFAULT_DIR)


class DownloadApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.download_main)
        self.command("raw")(self.download_raw_cmd)
        self.command("assets")(self.download_assets_cmd)
        self.command("timeseries")(self.download_timeseries_cmd)
        self.command("events")(self.download_events_cmd)
        self.command("files")(self.download_files_cmd)
        self.command("hierarchy")(self.download_hierarchy_cmd)
        if Flags.EXTEND_DOWNLOAD.is_enabled():
            self.command("datapoints")(self.download_datapoints_cmd)
        self.command("instances")(self.download_instances_cmd)
        self.command("charts")(self.download_charts_cmd)
        self.command("canvas")(self.download_canvas_cmd)

    @staticmethod
    def download_main(ctx: typer.Context) -> None:
        """Commands to download data from CDF into a temporary directory."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf download --help[/] for more information.")
        return None

    @staticmethod
    def download_raw_cmd(
        ctx: typer.Context,
        tables: Annotated[
            list[str] | None,
            typer.Argument(
                help="List of tables to download. If not provided, an interactive selection will be made.",
            ),
        ] = None,
        database: Annotated[
            str | None,
            typer.Option(
                "--database",
                "-d",
                help="Database to download from. If not provided, the user will be prompted to select a database.",
            ),
        ] = None,
        file_format: Annotated[
            RawFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the raw tables in. Supported formats: ndjson, yaml",
            ),
        ] = RawFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the raw tables. Supported formats: gzip, none.",
            ),
        ] = CompressionFormat.gzip,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the raw tables.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum the number of records to download from each table. Use -1 to download all records.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download RAW tables from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = DownloadCommand(client=client)

        if tables and database:
            selected_tables = [RawTableId(db_name=database, name=table) for table in tables]
        elif tables and not database:
            raise typer.BadParameter(
                "The '--database' option is required when specifying tables as arguments.",
                param_hint="--database",
            )
        elif not tables and database:
            selected_tables = RawTableInteractiveSelect(client, "download").select_tables(database=database)
        else:
            selected_tables = RawTableInteractiveSelect(client, "download").select_tables()

        cmd.run(
            lambda: cmd.download(
                selectors=[
                    RawTableSelector(
                        table=SelectedTable(db_name=item.db_name, table_name=item.name),
                        download_dir_name=item.db_name,
                    )
                    for item in selected_tables
                ],
                io=RawIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    def download_assets_cmd(
        self,
        ctx: typer.Context,
        data_sets: Annotated[
            list[str] | None,
            typer.Option(
                "--data-set",
                "-d",
                help="List of data sets to download assets from. If this is not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            AssetCentricFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the assets in.",
            ),
        ] = AssetCentricFormats.csv,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the assets.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the assets.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of assets to download from each dataset/hierarchy. Use -1 to download all assets.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download assets from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_sets is None:
            data_sets, file_format, compression, output_dir, limit = self._asset_centric_interactive(
                AssetInteractiveSelect(client, "download"),
                file_format,
                compression,
                output_dir,
                limit,
                "assets",
            )

        selectors = [
            DataSetSelector(
                kind="Assets",
                data_set_external_id=data_set,
                download_dir_name="asset-centric-assets",
            )
            for data_set in data_sets
        ]
        cmd = DownloadCommand(client=client)
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=AssetIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @classmethod
    def _asset_centric_interactive(
        cls,
        selector: AssetCentricInteractiveSelect,
        file_format: AssetCentricFormats,
        compression: CompressionFormat,
        output_dir: Path,
        limit: int,
        display_name: str,
        max_limit: int | None = None,
        available_formats: type[Enum] = AssetCentricFormats,
    ) -> tuple[list[str], AssetCentricFormats, CompressionFormat, Path, int]:
        data_sets = selector.select_data_sets()
        file_format = questionary.select(
            f"Select format to download the {display_name} in:",
            choices=[Choice(title=format_.value, value=format_) for format_ in available_formats],
            default=file_format,
        ).unsafe_ask()
        compression = questionary.select(
            f"Select compression format to use when downloading the {display_name}:",
            choices=[Choice(title=comp.value, value=comp) for comp in CompressionFormat],
            default=compression,
        ).unsafe_ask()
        output_dir = Path(
            questionary.path(
                f"Where to download the {display_name}:",
                default=str(output_dir),
                only_directories=True,
            ).unsafe_ask()
        )
        limit = int(
            questionary.text(
                f"The maximum number of {display_name} to download from each dataset. Use -1 to download all {display_name}.",
                default=str(limit),
                validate=lambda value: value.lstrip("-").isdigit() and (max_limit is None or int(value) <= max_limit),
            ).unsafe_ask()
        )
        return data_sets, file_format, compression, output_dir, limit

    def download_timeseries_cmd(
        self,
        ctx: typer.Context,
        data_sets: Annotated[
            list[str] | None,
            typer.Option(
                "--data-set",
                "-d",
                help="List of data sets to download time series from. If this is not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            AssetCentricFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the time series in.",
            ),
        ] = AssetCentricFormats.csv,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the time series.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the time series.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of time series to download from each dataset. Use -1 to download all time series.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download time series from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_sets is None:
            data_sets, file_format, compression, output_dir, limit = self._asset_centric_interactive(
                TimeSeriesInteractiveSelect(client, "download"),
                file_format,
                compression,
                output_dir,
                limit,
                "time series",
            )

        selectors = [
            DataSetSelector(
                kind="TimeSeries", data_set_external_id=data_set, download_dir_name="asset-centric-timeseries"
            )
            for data_set in data_sets
        ]
        cmd = DownloadCommand(client=client)
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=TimeSeriesIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    def download_events_cmd(
        self,
        ctx: typer.Context,
        data_sets: Annotated[
            list[str] | None,
            typer.Option(
                "--data-set",
                "-d",
                help="List of data sets to download events from. If this is not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            AssetCentricFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the events in.",
            ),
        ] = AssetCentricFormats.csv,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the events.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the events.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of events to download from each dataset. Use -1 to download all events.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download events from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_sets is None:
            data_sets, file_format, compression, output_dir, limit = self._asset_centric_interactive(
                EventInteractiveSelect(client, "download"),
                file_format,
                compression,
                output_dir,
                limit,
                "events",
            )

        selectors = [
            DataSetSelector(kind="Events", data_set_external_id=data_set, download_dir_name="asset-centric-events")
            for data_set in data_sets
        ]
        cmd = DownloadCommand(client=client)

        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=EventIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    def download_files_cmd(
        self,
        ctx: typer.Context,
        data_sets: Annotated[
            list[str] | None,
            typer.Option(
                "--data-set",
                "-d",
                help="List of data sets to download file metadata from. If this is not provided, an interactive selection will be made.",
            ),
        ] = None,
        include_file_contents: Annotated[
            bool,
            typer.Option(
                "--include-file-contents",
                "-c",
                help="Whether to include file contents when downloading assets. Note if you enable this option, you can"
                "only download 1000 files at a time.",
                hidden=not Flags.EXTEND_DOWNLOAD.is_enabled(),
            ),
        ] = False,
        file_format: Annotated[
            AssetCentricFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the file metadata in.",
            ),
        ] = AssetCentricFormats.csv,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the file metadata.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the file metadata.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of file metadata to download from each dataset. Use -1 to download all file metadata.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download file metadata from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_sets is None:
            if Flags.EXTEND_DOWNLOAD.is_enabled():
                include_file_contents = questionary.select(
                    "Do you want to include file contents when downloading file metadata?",
                    choices=[
                        Choice(title="Yes", value=True),
                        Choice(title="No", value=False),
                    ],
                ).unsafe_ask()
            else:
                include_file_contents = False

            available_formats = FileContentFormats if include_file_contents else AssetCentricFormats
            file_format = FileContentFormats.ndjson if include_file_contents else file_format  # type: ignore[assignment]
            data_sets, file_format, compression, output_dir, limit = self._asset_centric_interactive(
                FileMetadataInteractiveSelect(client, "download"),
                file_format,
                compression,
                output_dir,
                limit if not include_file_contents else 1000,
                "file metadata",
                max_limit=1000 if include_file_contents else None,
                available_formats=available_formats,
            )

        io: StorageIO
        selectors: list[DataSelector]
        if include_file_contents:
            if limit == -1 or limit > 1000:
                limit = 1000
                print(
                    "[yellow]When including file contents, the maximum number of files that can be downloaded at a time is 1000. "
                )
            if file_format == AssetCentricFormats.csv or file_format == AssetCentricFormats.parquet:
                print(
                    "[red]When including file contents, the only supported format is ndjson. Overriding the format to ndjson.[/]"
                )
                file_format = AssetCentricFormats.ndjson
            files = client.files.list(data_set_external_ids=data_sets, limit=limit)
            download_dir_name = "asset-centric-files"
            selector = FileIdentifierSelector(
                identifiers=tuple([FileInternalID(internal_id=file.id) for file in files]),
                download_dir_name=download_dir_name,
            )
            selectors = [selector]
            io = FileContentIO(client, output_dir / download_dir_name)
        else:
            selectors = [
                DataSetSelector(
                    kind="FileMetadata", data_set_external_id=data_set, download_dir_name="asset-centric-files"
                )
                for data_set in data_sets
            ]
            io = FileMetadataIO(client)

        cmd = DownloadCommand(client=client)
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,  # type: ignore[misc]
                io=io,
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_hierarchy_cmd(
        ctx: typer.Context,
        hierarchy: Annotated[
            str | None,
            typer.Argument(
                help="The asset hierarchy to download.",
            ),
        ] = None,
        file_format: Annotated[
            HierarchyFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format for downloading the asset hierarchy.",
            ),
        ] = HierarchyFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the assets.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the asset hierarchy.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of resources to download for each type. Use -1 to download all assets.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download an asset hierarchy from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = DownloadCommand(client=client)

        if hierarchy is None:
            selector = AssetInteractiveSelect(client, "download")
            hierarchy = selector.select_hierarchy(allow_empty=False)

        selectors = [
            # MyPy cannot see that resource_type is one of the allowed literals.
            AssetSubtreeSelector(hierarchy=hierarchy, kind=resource_type, download_dir_name=f"hierarchy-{hierarchy}")  # type: ignore[arg-type]
            for resource_type in ["Assets", "Events", "FileMetadata", "TimeSeries"]
        ]
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=HierarchyIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @classmethod
    def download_instances_cmd(
        cls,
        ctx: typer.Context,
        instance_space: Annotated[
            str | None,
            typer.Option(
                "--instance-space",
                "-s",
                help="The instance space to download instances from. If not provided, an interactive "
                "selection will be made.",
            ),
        ] = None,
        schema_space: Annotated[
            str | None,
            typer.Option(
                "--schema-space",
                "-c",
                help="The schema space where the views are located.",
            ),
        ] = None,
        view_external_ids: Annotated[
            list[str] | None,
            typer.Option(
                "--view",
                "-w",
                help="List of view external IDs to download properties for the "
                "instances. To specify version use a forward slash, e.g. viewExternalId/v1.",
            ),
        ] = None,
        instance_type: Annotated[
            InstanceTypes,
            typer.Option(
                "--instance-type",
                "-t",
                help="The type of instances to download.",
            ),
        ] = InstanceTypes.node,
        file_format: Annotated[
            InstanceFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the instances in.",
            ),
        ] = InstanceFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the instances.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the instances.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum the number of instances to download from each view. Use -1 to download all.",
            ),
        ] = 10_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download Instances from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = DownloadCommand(client=client)

        selectors: list[InstanceSelector]
        if instance_space is None:
            selector = DataModelingSelect(client, "download instances")
            data_model = selector.select_data_model(
                inline_views=False,
                message="Select the data model through which to download instances:",
                include_global=True,
            )
            selected_views = selector.select_view(
                multiselect=True,
                message="Select views to download instance properties from.",
                filter=ViewSelectFilter(
                    strategy="dataModel",
                    include_global=True,
                    data_model=data_model.as_id(),
                ),
            )
            select_instance_space = questionary.confirm(
                "Do you want to select an instance space to download from? If no, all instances from the selected views will be downloaded.",
                default=False,
            ).unsafe_ask()
            instance_spaces: tuple[str, ...] | None = None
            if select_instance_space:
                instance_spaces = tuple(selector.select_instance_space(multiselect=True))
            selectors = []
            download_dir_name = sanitize_filename(data_model.external_id)
            for view in selected_views:
                view_instance_type = selector.select_instance_type(
                    view.used_for,
                    message=f"Select instance type to download for view {view.space}:{view.external_id}(version={view.version})",
                )
                selectors.append(
                    InstanceViewSelector(
                        view=SelectedView(
                            space=view.space,
                            external_id=view.external_id,
                            version=view.version,
                        ),
                        instance_spaces=instance_spaces,
                        instance_type=view_instance_type,
                        download_dir_name=download_dir_name,
                    )
                )
            output_dir, file_format, compression, limit = cls._interactive_select_shared(  # type: ignore[assignment]
                output_dir, file_format, InstanceFormats, compression, limit, "instances", "view"
            )
        elif schema_space is None and view_external_ids is None:
            selectors = [
                InstanceSpaceSelector(
                    instance_space=instance_space,
                    instance_type=instance_type.value,
                    download_dir_name=instance_space,
                )
            ]
        elif schema_space is not None and view_external_ids is not None:
            selectors = [
                InstanceSpaceSelector(
                    instance_space=instance_space,
                    view=SelectedView(
                        space=schema_space,
                        external_id=view_id_str.split("/", maxsplit=1)[0],
                        version=view_id_str.split("/", maxsplit=1)[1] if "/" in view_id_str else None,
                    ),
                    instance_type=instance_type.value,
                    download_dir_name=instance_space,
                )
                for view_id_str in view_external_ids
            ]
        else:
            raise typer.BadParameter(
                "Both '--schema-space' and '--view' must be provided together.",
                param_hint="--view",
            )

        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=InstanceIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @classmethod
    def _interactive_select_shared(
        cls,
        output_dir: Path,
        file_format: Enum,
        file_format_options: type[Enum],
        compression: CompressionFormat,
        limit: int,
        display_name: str,
        selector_type: str,
    ) -> tuple[Path, Enum, Enum, int]:
        """Interactive selection of output_dir, file_format, compression and limit for the download commands."""
        selected_output_dir = Path(
            questionary.path("Where to download the data:", default=str(output_dir), only_directories=True).unsafe_ask()
        )

        file_formats = [Choice(title=format_.value, value=format_) for format_ in file_format_options]
        if len(file_formats) == 1:
            selected_file_format = file_formats[0].value
        else:
            selected_file_format = questionary.select(
                "Select format to download the data in:",
                choices=file_formats,
                default=file_format,  # type: ignore[arg-type]
            ).unsafe_ask()

        selected_compression = questionary.select(
            "Select compression format to use when downloading the data:",
            choices=[Choice(title=comp.value, value=comp) for comp in CompressionFormat],
            default=compression,
        ).unsafe_ask()
        selected_limit = int(
            questionary.text(
                f"The maximum number of {display_name} to download per {selector_type}. Use -1 to download all {display_name}.",
                default=str(limit),
                validate=lambda value: value.lstrip("-").isdigit() and (int(value) == -1 or int(value) > 0),
            ).unsafe_ask()
        )
        return selected_output_dir, selected_file_format, selected_compression, selected_limit  # type: ignore[return-value]

    @staticmethod
    def download_datapoints_cmd(
        dataset: Annotated[
            str | None,
            typer.Argument(
                help="The dataset to download timeseries from. If not provided, an interactive selection will be made.",
            ),
        ] = None,
        start_time: Annotated[
            str | None,
            typer.Option(
                "--start-time",
                "-s",
                help="The start time for the datapoints to download. Can be in RFC3339 format or as a relative time (e.g., '1d-ago'). If not provided, all datapoints from the beginning will be downloaded.",
            ),
        ] = None,
        end_time: Annotated[
            str | None,
            typer.Option(
                "--end-time",
                "-e",
                help="The end time for the datapoints to download. Can be in RFC3339 format or as a relative time (e.g., '1d-ago'). If not provided, all datapoints up to the latest will be downloaded.",
            ),
        ] = None,
        datapoint_type: Annotated[
            DatapointsDataTypes,
            typer.Option(
                "--data-type",
                "-d",
                help="The type of datapoints to download.",
            ),
        ] = DatapointsDataTypes.numeric,
        file_format: Annotated[
            DatapointFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format for downloading the datapoints.",
            ),
        ] = DatapointFormats.csv,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the datapoints.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of timeseries to download datapoints from. Use -1 to download all timeseries."
                "The maximum number of datapoints in total is 10 million and 100 000 per timeseries.",
                max=10_000_000,
            ),
        ] = 1000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download Datapoints from CDF into a temporary ."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if dataset is None:
            interactive = TimeSeriesInteractiveSelect(client, "download datapoints")
            dataset = interactive.select_data_set(allow_empty=False)

            datapoint_type = questionary.select(
                "Select the type of datapoints to download:",
                choices=[Choice(title=dt.value, value=dt) for dt in DatapointsDataTypes],
                default=datapoint_type,
            ).unsafe_ask()

            start_time = (
                questionary.text(
                    "Enter the start time for the datapoints to download (RFC3339 format or relative time, e.g., '1d-ago'). Leave empty to download from the beginning.",
                    default=start_time or "",
                ).unsafe_ask()
                or None
            )
            end_time = (
                questionary.text(
                    "Enter the end time for the datapoints to download (RFC3339 format or relative time, e.g., '1d-ago'). Leave empty to download up to the latest.",
                    default=end_time or "",
                ).unsafe_ask()
                or None
            )
            file_format = questionary.select(
                "Select format to download the datapoints in:",
                choices=[Choice(title=format_.value, value=format_) for format_ in DatapointFormats],
                default=file_format,
            ).unsafe_ask()
            output_dir = Path(
                questionary.path(
                    "Where to download the datapoints:", default=str(output_dir), only_directories=True
                ).unsafe_ask()
            )
            limit = int(
                questionary.text(
                    "The maximum number of timeseries to download datapoints from. Use -1 to download all timeseries."
                    "The maximum number of datapoints in total is 10 million and 100 000 per timeseries.",
                    default=str(limit),
                    validate=lambda value: value.lstrip("-").isdigit() and (int(value) == -1 or int(value) > 0),
                ).unsafe_ask()
            )

        cmd = DownloadCommand(client=client)
        selector = DataPointsDataSetSelector(
            data_set_external_id=dataset,
            start=start_time,
            end=end_time,
            data_type=datapoint_type.value,
            download_dir_name=f"dataset-{dataset}",
        )
        cmd.run(
            lambda: cmd.download(
                selectors=[selector],
                io=DatapointsIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression="none",
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_charts_cmd(
        ctx: typer.Context,
        external_ids: Annotated[
            list[str] | None,
            typer.Argument(
                help="List of chart external IDs to download. If not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            ChartFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format for downloading the charts.",
            ),
        ] = ChartFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the instances.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the charts.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of charts to download. Use -1 to download all charts.",
            ),
        ] = 1000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download Charts from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = DownloadCommand(client=client)
        selector: ChartSelector
        if external_ids is None:
            selected_external_ids = InteractiveChartSelect(client).select_external_ids()
            selector = ChartExternalIdSelector(external_ids=tuple(selected_external_ids), download_dir_name="charts")
        else:
            selector = ChartExternalIdSelector(external_ids=tuple(external_ids), download_dir_name="charts")

        cmd.run(
            lambda: cmd.download(
                selectors=[selector],
                io=ChartIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_canvas_cmd(
        ctx: typer.Context,
        external_ids: Annotated[
            list[str] | None,
            typer.Argument(
                help="List of canvas external IDs to download. If not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            CanvasFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format for downloading the canvas.",
            ),
        ] = CanvasFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the canvas.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the canvas.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of canvas to download. Use -1 to download all canvas.",
            ),
        ] = 1000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download Canvas from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = DownloadCommand(client=client)
        selector: CanvasSelector
        if external_ids is None:
            selected_external_ids = InteractiveCanvasSelect(client).select_external_ids()
            selector = CanvasExternalIdSelector(external_ids=tuple(selected_external_ids), download_dir_name="canvas")
        else:
            selector = CanvasExternalIdSelector(external_ids=tuple(external_ids), download_dir_name="canvas")

        cmd.run(
            lambda: cmd.download(
                selectors=[selector],
                io=CanvasIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )



"""Data Product Versions API for managing versions of CDF data products."""

from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_product_version import (
    DataProductVersionRequest,
    DataProductVersionResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import DataProductVersionId


class DataProductVersionsAPI(CDFResourceAPI[DataProductVersionResponse]):
    """API for managing data product versions.

    All endpoints are scoped under a parent data product: /dataproducts/{externalId}/versions.
    Versions are identified by a user-specified semantic version string (major.minor.patch).
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/dataproducts/{externalId}/versions", item_limit=1),
                "retrieve": Endpoint(method="GET", path="/dataproducts/{externalId}/versions/{version}", item_limit=1),
                "update": Endpoint(method="POST", path="/dataproducts/{externalId}/versions/update", item_limit=1),
                "delete": Endpoint(method="POST", path="/dataproducts/{externalId}/versions/delete", item_limit=10),
                "list": Endpoint(method="GET", path="/dataproducts/{externalId}/versions", item_limit=10),
            },
            api_version="alpha",
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[DataProductVersionResponse]:
        return PagedResponse[DataProductVersionResponse].model_validate_json(response.body)

    @staticmethod
    def _group_by_parent(
        items: Sequence[DataProductVersionRequest],
    ) -> dict[str, list[DataProductVersionRequest]]:
        by_parent: dict[str, list[DataProductVersionRequest]] = {}
        for item in items:
            by_parent.setdefault(item.data_product_external_id, []).append(item)
        return by_parent

    def create(self, items: Sequence[DataProductVersionRequest]) -> list[DataProductVersionResponse]:
        results: list[DataProductVersionResponse] = []
        for dp_ext_id, group in self._group_by_parent(items).items():
            url = self._make_url(self._method_endpoint_map["create"].path.format(externalId=dp_ext_id))
            for item in group:
                response = self._http_client.request_single_retries(
                    RequestMessage(
                        endpoint_url=url,
                        method="POST",
                        body_content={"items": [item.dump()]},
                        api_version=self._api_version,
                    )
                )
                page = self._validate_page_response(response.get_success_or_raise())
                for version in page.items:
                    version.data_product_external_id = dp_ext_id
                results.extend(page.items)
        return results

    def retrieve(
        self,
        items: Sequence[DataProductVersionId],
        ignore_unknown_ids: bool = False,
    ) -> list[DataProductVersionResponse]:
        results: list[DataProductVersionResponse] = []
        for item in items:
            url = self._make_url(
                self._method_endpoint_map["retrieve"].path.format(
                    externalId=item.data_product_external_id, version=item.version
                )
            )
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=url,
                    method="GET",
                    api_version=self._api_version,
                )
            )
            if isinstance(response, SuccessResponse):
                ver = DataProductVersionResponse.model_validate_json(response.body)
                ver.data_product_external_id = item.data_product_external_id
                results.append(ver)
            elif ignore_unknown_ids:
                continue
            else:
                _ = response.get_success_or_raise()
        return results

    def update(
        self,
        items: Sequence[DataProductVersionRequest],
        mode: Literal["patch", "replace"] = "replace",
    ) -> list[DataProductVersionResponse]:
        results: list[DataProductVersionResponse] = []
        for dp_ext_id, group in self._group_by_parent(items).items():
            url = self._make_url(self._method_endpoint_map["update"].path.format(externalId=dp_ext_id))
            for item in group:
                response = self._http_client.request_single_retries(
                    RequestMessage(
                        endpoint_url=url,
                        method="POST",
                        body_content={"items": [item.as_update(mode=mode)]},
                        api_version=self._api_version,
                    )
                )
                page = self._validate_page_response(response.get_success_or_raise())
                for ver in page.items:
                    ver.data_product_external_id = dp_ext_id
                results.extend(page.items)
        return results

    def delete(self, ids: Sequence[DataProductVersionId]) -> None:
        by_parent: defaultdict[str, list[str]] = defaultdict(list)
        for id_ in ids:
            by_parent[id_.data_product_external_id].append(id_.version)
        for dp_ext_id, versions in by_parent.items():
            url = self._make_url(self._method_endpoint_map["delete"].path.format(externalId=dp_ext_id))
            self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=url,
                    method="POST",
                    body_content={"items": [{"version": v} for v in versions]},
                    api_version=self._api_version,
                )
            ).get_success_or_raise()

    def iterate(
        self, data_product_external_id: str, limit: int | None = 10
    ) -> Iterable[list[DataProductVersionResponse]]:
        path = self._method_endpoint_map["list"].path.format(externalId=data_product_external_id)
        for batch in self._iterate(limit=limit, endpoint_path=path):
            for item in batch:
                item.data_product_external_id = data_product_external_id
            yield batch

    def list(self, data_product_external_id: str, limit: int | None = 10) -> list[DataProductVersionResponse]:
        path = self._method_endpoint_map["list"].path.format(externalId=data_product_external_id)
        items = self._list(limit=limit, endpoint_path=path)
        for item in items:
            item.data_product_external_id = data_product_external_id
        return items



UPSERT_BODY = """
mutation UpsertGraphQlDmlVersion($dmCreate: GraphQlDmlVersionUpsert!) {
    upsertGraphQlDmlVersion(graphQlDmlVersion: $dmCreate) {
        errors {
            kind
            message
            hint
            location {
                start {
                    line
                    column
                }
            }
        }
        result {
            space
            externalId
            version
            name
            description
            graphQlDml
            createdTime
            lastUpdatedTime
        }
    }
}
"""



import sys
from collections.abc import Iterator
from typing import Literal

from pydantic import RootModel, model_validator

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class UserProfilesConfiguration(BaseModelObject):
    enabled: bool


class Claim(BaseModelObject):
    claim_name: str


class OidcConfiguration(BaseModelObject):
    jwks_url: str
    token_url: str | None = None
    issuer: str
    audience: str
    skew_ms: int | None = None
    access_claims: list[Claim]
    scope_claims: list[Claim]
    log_claims: list[Claim]
    is_group_callback_enabled: bool | None = None
    identity_provider_scope: str | None = None


class OrganizationResponse(BaseModelObject):
    name: str
    url_name: str
    organization: str
    user_profiles_configuration: UserProfilesConfiguration
    oidc_configuration: OidcConfiguration | None = None


class ProjectStatus(BaseModelObject):
    """Project status information."""

    data_modeling_status: Literal["HYBRID", "DATA_MODELING_ONLY"]
    url_name: str


class ProjectStatusList(RootModel[list[ProjectStatus]]):
    root: list[ProjectStatus]
    _project: str

    def __iter__(self) -> Iterator[ProjectStatus]:  # type: ignore[override]
        return iter(self.root)

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, index: int) -> ProjectStatus:
        return self.root[index]

    @classmethod
    def _load(cls, data: list[dict]) -> Self:
        """Load from a list of dictionaries."""
        return cls(root=[ProjectStatus._load(item) for item in data], _project="")

    @property
    def this_project(self) -> ProjectStatus:
        """Returns the ProjectStatus of the current project."""
        for item in self.root:
            if self._project == item.url_name:
                return item
        raise ValueError(f"Project '{self._project}' not found in the list of projects.")

    @model_validator(mode="before")
    def move_item_to_root(cls, data: dict) -> dict:
        """Move the list of projects from the 'projects' key to the root."""
        if "items" in data:
            return data["items"]
        return data



"""Agents API for managing CDF AI agents.

Based on the API specification at:
https://api-docs.cognite.com/20230101-alpha/tag/Agents/operation/main_ai_agents_post

Note: This is an alpha API and may change in future releases.
"""

from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.agent import AgentRequest, AgentResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class AgentsAPI(CDFResourceAPI[AgentResponse]):
    """API for managing CDF AI agents.

    Note: This is an alpha API and may change in future releases.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/ai/agents", item_limit=1),
                "retrieve": Endpoint(method="POST", path="/ai/agents/byids", item_limit=100),
                "delete": Endpoint(method="POST", path="/ai/agents/delete", item_limit=1),
                "list": Endpoint(method="GET", path="/ai/agents", item_limit=1000),
            },
            api_version="beta",
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[AgentResponse]:
        return PagedResponse[AgentResponse].model_validate_json(response.body)

    def create(self, items: Sequence[AgentRequest]) -> list[AgentResponse]:
        """Apply (create or update) agents in CDF.

        Args:
            items: List of AgentRequest objects to apply.

        Returns:
            List of applied AgentResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def update(self, items: Sequence[AgentRequest]) -> list[AgentResponse]:
        """Update agents in CDF.

        Args:
            items: List of AgentRequest objects to update.
        Returns:
            List of updated AgentResponse objects.
        """
        # Implemented as an alias to create (upsert) to have a standardized interface.
        return self._request_item_response(items, "upsert")

    def retrieve(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> list[AgentResponse]:
        """Retrieve agents from CDF by external ID.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of retrieved AgentResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete agents from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def iterate(
        self,
        limit: int | None = None,
    ) -> Iterable[list[AgentResponse]]:
        """Iterate over all agents in CDF.

        Args:
            limit: Maximum total number of agents to return.

        Returns:
            Iterable of lists of AgentResponse objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = None) -> list[AgentResponse]:
        """List all agents in CDF.

        Args:
            limit: Maximum total number of agents to return.

        Returns:
            List of AgentResponse objects.
        """
        return self._list(limit=limit)



import hashlib
from abc import ABC
from typing import Literal

from ._base import DataSelector


class ChartSelector(DataSelector, ABC):
    kind: Literal["Charts"] = "Charts"


class ChartOwnerSelector(ChartSelector):
    type: Literal["chartOwner"] = "chartOwner"
    owner_id: str

    def __str__(self) -> str:
        return self.owner_id


class AllChartsSelector(ChartSelector):
    type: Literal["allCharts"] = "allCharts"

    def __str__(self) -> str:
        return "all"


class ChartExternalIdSelector(ChartSelector):
    type: Literal["chartExternalId"] = "chartExternalId"
    external_ids: tuple[str, ...]

    def __str__(self) -> str:
        hash_ = hashlib.md5(",".join(sorted(self.external_ids)).encode()).hexdigest()[:8]
        return f"chart_count_{len(self.external_ids)}_hash_{hash_}"



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.workflow import WorkflowRequest, WorkflowResponse

from .workflow_triggers import WorkflowTriggersAPI
from .workflow_versions import WorkflowVersionsAPI


class WorkflowsAPI(CDFResourceAPI[WorkflowResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/workflows", item_limit=1),
                "retrieve": Endpoint(method="GET", path="/workflows/{workflowExternalId}", item_limit=1),
                "delete": Endpoint(method="POST", path="/workflows/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/workflows", item_limit=100),
            },
        )
        self.versions = WorkflowVersionsAPI(http_client)
        self.triggers = WorkflowTriggersAPI(http_client)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[WorkflowResponse]:
        return PagedResponse[WorkflowResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[WorkflowRequest]) -> list[WorkflowResponse]:
        """Create or update workflows in CDF.

        Args:
            items: List of WorkflowRequest objects to create or update.
        Returns:
            List of created/updated WorkflowResponse objects.
        """
        return self._request_item_response(items, "upsert")

    # This is a duplicate of the create method, included to standardize the API interface.
    def update(self, items: Sequence[WorkflowRequest]) -> list[WorkflowResponse]:
        """Create or update workflows in CDF.

        Args:
            items: List of WorkflowRequest objects to create or update.
        Returns:
            List of created/updated WorkflowResponse objects.
        """
        return self.create(items)

    def retrieve(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> list[WorkflowResponse]:
        """Retrieve workflows from CDF by external ID.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of retrieved WorkflowResponse objects.
        """
        result: list[WorkflowResponse] = []
        endpoint = self._method_endpoint_map["retrieve"]
        for item in items:
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=self._make_url(f"/workflows/{item.external_id}"),
                    method=endpoint.method,
                )
            )
            if isinstance(response, SuccessResponse):
                result.append(WorkflowResponse.model_validate_json(response.body))
            elif ignore_unknown_ids:
                continue
            else:
                _ = response.get_success_or_raise()
        return result

    def delete(self, items: Sequence[ExternalId]) -> None:
        """Delete workflows from CDF.

        Args:
            items: List of ExternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[WorkflowResponse]:
        """Iterate over all workflows in CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of WorkflowResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(
        self,
        limit: int = 100,
    ) -> Iterable[list[WorkflowResponse]]:
        """Iterate over all workflows in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of WorkflowResponse objects.
        """
        return self._iterate(limit=limit)

    def list(
        self,
        limit: int | None = 100,
    ) -> list[WorkflowResponse]:
        """List all workflows in CDF.

        Returns:
            List of WorkflowResponse objects.
        """
        return self._list(limit=limit)



from typing import cast

from cognite.client import CogniteClient
from rich.console import Console

from cognite_toolkit._cdf_tk.client.api.charts import ChartsAPI
from cognite_toolkit._cdf_tk.client.api.legacy.canvas import CanvasAPI
from cognite_toolkit._cdf_tk.client.api.location_filters import LocationFiltersAPI
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient

from .api.agents import AgentsAPI
from .api.annotations import AnnotationsAPI
from .api.assets import AssetsAPI
from .api.cognite_files import CogniteFilesAPI
from .api.containers import ContainersAPI
from .api.data_models import DataModelsAPI
from .api.data_products import DataProductsAPI
from .api.datapoint_subscription import DatapointSubscriptionsAPI
from .api.datasets import DataSetsAPI
from .api.events import EventsAPI
from .api.extraction_pipelines import ExtractionPipelinesAPI
from .api.filemetadata import FileMetadataAPI
from .api.functions import FunctionsAPI
from .api.graphql_data_models import GraphQLDataModelsAPI
from .api.groups import GroupsAPI
from .api.hosted_extractors import HostedExtractorsAPI
from .api.infield import InfieldAPI
from .api.instances import InstancesAPI
from .api.labels import LabelsAPI
from .api.lookup import LookUpGroup
from .api.migration import MigrationAPI
from .api.principals import PrincipalsAPI
from .api.project import ProjectAPI
from .api.raw import RawAPI
from .api.relationships import RelationshipsAPI
from .api.robotics import RoboticsAPI
from .api.search_config import SearchConfigurationsAPI
from .api.security_categories import SecurityCategoriesAPI
from .api.sequences import SequencesAPI
from .api.simulators import SimulatorsAPI
from .api.spaces import SpacesAPI
from .api.streamlit_ import StreamlitAPI
from .api.streams import StreamsAPI
from .api.three_d import ThreeDAPI
from .api.timeseries import TimeSeriesAPI
from .api.token import TokenAPI
from .api.transformations import TransformationsAPI
from .api.verify import VerifyAPI
from .api.views import ViewsAPI
from .api.workflows import WorkflowsAPI
from .config import ToolkitClientConfig


class ToolAPI:
    """This is reimplemented CogniteAPIs in Toolkit"""

    def __init__(self, http_client: HTTPClient, console: Console) -> None:
        self.http_client = http_client
        self.agents = AgentsAPI(http_client)
        self.annotations = AnnotationsAPI(http_client)
        self.assets = AssetsAPI(http_client)
        self.cognite_files = CogniteFilesAPI(http_client)
        self.datasets = DataSetsAPI(http_client)
        self.datapoint_subscriptions = DatapointSubscriptionsAPI(http_client)
        self.events = EventsAPI(http_client)
        self.extraction_pipelines = ExtractionPipelinesAPI(http_client)
        self.functions = FunctionsAPI(http_client)
        self.groups = GroupsAPI(http_client)
        self.hosted_extractors = HostedExtractorsAPI(http_client)
        self.instances = InstancesAPI(http_client)
        self.spaces = SpacesAPI(http_client)
        self.views = ViewsAPI(http_client)
        self.containers = ContainersAPI(http_client)
        self.data_models = DataModelsAPI(http_client)
        self.graphql_data_models = GraphQLDataModelsAPI(http_client)
        self.labels = LabelsAPI(http_client)
        self.location_filters = LocationFiltersAPI(http_client)
        self.filemetadata = FileMetadataAPI(http_client)
        self.raw = RawAPI(http_client)
        self.robotics = RoboticsAPI(http_client)
        self.security_categories = SecurityCategoriesAPI(http_client)
        self.relationships = RelationshipsAPI(http_client)
        self.sequences = SequencesAPI(http_client)
        self.search_configurations = SearchConfigurationsAPI(http_client)
        self.simulators = SimulatorsAPI(http_client)
        self.three_d = ThreeDAPI(http_client)
        self.timeseries = TimeSeriesAPI(http_client)
        self.transformations = TransformationsAPI(http_client)
        self.workflows = WorkflowsAPI(http_client)
        self.data_products = DataProductsAPI(http_client)
        self.streamlit = StreamlitAPI(http_client)


class ToolkitClient(CogniteClient):
    def __init__(
        self,
        config: ToolkitClientConfig | None = None,
        console: Console | None = None,
    ) -> None:
        super().__init__(config=config)
        http_client = HTTPClient(self.config, console=console)
        self.http_client = http_client
        self.console = console or Console(markup=True)
        self.tool = ToolAPI(http_client, self.console)

        self.verify = VerifyAPI(self._config, self._API_VERSION, self)
        self.lookup = LookUpGroup(self._config, self._API_VERSION, self, self.console)
        self.canvas = CanvasAPI(self.data_modeling.instances)
        self.migration = MigrationAPI(self.data_modeling.instances, http_client)
        self.token = TokenAPI(self)
        self.charts = ChartsAPI(http_client)
        self.project = ProjectAPI(http_client)
        self.principals = PrincipalsAPI(http_client=http_client, project_api=self.project)
        self.infield = InfieldAPI(http_client)
        self.streams = StreamsAPI(http_client)

    @property
    def config(self) -> ToolkitClientConfig:
        """Returns a config object containing the configuration for the current client.

        Returns:
            ToolkitClientConfig: The configuration object.
        """
        return cast(ToolkitClientConfig, self._config)
